# -*- coding: utf-8 -*-

# @Time    : 2019/11/1 13:18
# @Email   : 986798607@qq.ele_ratio
# @Software: PyCharm
# @License: BSD 3-Clause

from abc import abstractmethod
from typing import List, Tuple, Union, Dict

import numpy as np
from pymatgen.core.composition import Composition as PMGComp
from pymatgen.core.periodic_table import get_el_sp
from pymatgen.util.string import formula_double_format

from featurebox.featurizers.atom.mapper import AtomTableMap, BinaryMap
from featurebox.featurizers.base_transform import BaseFeature
from featurebox.featurizers.extrastats import PropertyStats


class BaseCompositionFeature(BaseFeature):
    """
    BaseCompositionFeature is the basis for composition data.
    the subclass should be re-implemented, such as:
    ::

        def mix_function(self, elems:List, nums:List):
            w_ = np.array(nums)
            return w_.dot(elems)

    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        """
        Base class for composition feature.
        """
        super().__init__(n_jobs, on_errors=on_errors, return_type=return_type)
        if data_map is None:
            data_map = AtomTableMap(tablename="oe.csv", search_tp="name")
        self.data_map = data_map
        # change
        self.data_map.weight = False
        self.data_map.n_jobs = 1
        self.data_map.search_tp = "name"

        self.search_tp = "name"
        self.convert = self.convert_dict

    def convert_dict(self, atoms: dict) -> np.ndarray:
        """
        Convert atom {symbol: fraction} list to numeric features
        """
        if isinstance(atoms, dict):
            atoms = [{k: v} for k, v in atoms.items()]

        numbers = np.array([list(ai.values())[0] for ai in atoms])

        ele = self.data_map.convert(atoms)
        if len(atoms)==1:
            ele = np.array(ele).reshape((len(atoms), -1))
        return self.mix_function(ele, numbers)

    @abstractmethod
    def mix_function(self, elems: List, nums: List):
        """

        Parameters
        ----------
        elems: list
            Elements in compound.
        nums: list
            Number of each element.

        Returns
        -------
        descriptor: numpy.ndarray
        """


class WeightedAverage(BaseCompositionFeature):
    """
    Examples
    ---------
    >>> from featurebox.featurizers.atom.mapper import AtomTableMap, AtomJsonMap
    >>> data_map = AtomJsonMap(search_tp="name", n_jobs=1)
    >>> wa = WeightedAverage(data_map, n_jobs=1,return_type="df")
    >>> x3 = [{"H": 2, "Pd": 1},{"He":1,"Al":4}]
    >>> wa.fit_transform(x3)
             0         1         2   ...        13        14        15
    0  0.422068  0.360958  0.201433  ... -0.459164 -0.064783 -0.250939
    1  0.007163 -0.471498 -0.072860  ...  0.206306 -0.041006  0.055843
    <BLANKLINE>
    [2 rows x 16 columns]

    >>> wa.set_feature_labels(["fea_{}".format(_) for _ in range(16)])
    >>> wa.fit_transform(x3)
          fea_0     fea_1     fea_2  ...    fea_13    fea_14    fea_15
    0  0.422068  0.360958  0.201433  ... -0.459164 -0.064783 -0.250939
    1  0.007163 -0.471498 -0.072860  ...  0.206306 -0.041006  0.055843
    <BLANKLINE>
    [2 rows x 16 columns]

    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        w_ = nums / np.sum(nums)
        return w_.dot(elems)


class WeightedSum(BaseCompositionFeature):
    """
    Examples
    --------
    >>> from featurebox.featurizers.atom.mapper import AtomTableMap, AtomJsonMap
    >>> data_map = AtomJsonMap(search_tp="name", n_jobs=1)
    >>> wa = WeightedSum(data_map, n_jobs=1,return_type="df")
    >>> x3 = [{"H": 2, "Pd": 1},{"He":1,"Al":4}]
    >>> wa.fit_transform(x3)
             0         1         2   ...        13        14        15
    0  1.266204  1.082873  0.604300  ... -1.377492 -0.194350 -0.752816
    1  0.035813 -2.357490 -0.364302  ...  1.031530 -0.205029  0.279215
    <BLANKLINE>
    [2 rows x 16 columns]

    >>> wa.set_feature_labels(["fea_{}".format(_) for _ in range(16)])
    >>> wa.fit_transform(x3)
          fea_0     fea_1     fea_2  ...    fea_13    fea_14    fea_15
    0  1.266204  1.082873  0.604300  ... -1.377492 -0.194350 -0.752816
    1  0.035813 -2.357490 -0.364302  ...  1.031530 -0.205029  0.279215
    <BLANKLINE>
    [2 rows x 16 columns]

    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        w_ = np.array(nums)
        return w_.dot(elems)


class GeometricMean(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems: np.ndarray, nums):
        w_ = np.array(nums).reshape(-1, 1)
        tmp = elems ** w_
        return np.power(tmp.prod(axis=0), 1 / sum(w_))


class HarmonicMean(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        w_ = np.array(nums)
        tmp = w_.dot(elems)
        return sum(w_) / tmp


class WeightedVariance(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems: np.ndarray, nums):
        w_ = nums / np.sum(nums)
        mean_ = w_.dot(elems)
        var_ = elems - mean_
        return w_.dot(var_ ** 2)


class MaxPooling(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        return np.max(elems)


class MinPooling(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map=data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        return np.min(elems)


class ExtraMix(BaseCompositionFeature):
    """
    See Also:
        :class:`WeightedSum`
    """

    def __init__(self, data_map: BinaryMap, stats: Tuple[str] = ("mean",), n_jobs: int = 1,
                 on_errors: str = 'raise', return_type: str = 'df'):
        super().__init__(data_map=data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.stats = stats

    def mix_function(self, elems, nums):
        all_attributes = []
        for stat in self.stats:
            all_attributes.append(PropertyStats.calc_stat(elems, stat, nums))

        return np.array(all_attributes).ravel()


class DepartElementFeature(BaseCompositionFeature):
    """
    Get the table of element data.

    Examples
    ----------
    >>> from featurebox.featurizers.atom.mapper import AtomTableMap, AtomJsonMap
    >>> data_map = AtomJsonMap(search_tp="name", n_jobs=1)
    >>> wa = DepartElementFeature(data_map,n_composition=2, n_jobs=1,return_type="df")
    >>> x3 = [{"H": 2, "Pd": 1},{"He":1,"Al":4}]
    >>> wa.set_feature_labels(["fea_{}".format(_) for _ in range(16)])
    >>> wa.fit_transform(x3)
        fea_0_0   fea_0_1   fea_1_0  ...  fea_14_1  fea_15_0  fea_15_1
    0  0.352363  0.561478  0.635952  ... -0.236541 -0.270104 -0.212607
    1 -0.067220  0.025758  0.141113  ... -0.092577 -0.042185  0.080350
    <BLANKLINE>
    [2 rows x 32 columns]

    """

    def __init__(self, data_map: BinaryMap, n_composition: int, n_jobs: int = 1, on_errors: str = 'raise',
                 return_type: str = 'df'):

        super().__init__(data_map=data_map, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.n_composition = n_composition

    def mix_function(self, elems: np.ndarray, nums=None):

        return elems.ravel(order='F')

    def convert(self, comp: Union[Dict, PMGComp]):
        elems_, nums_ = [], []
        if isinstance(comp, PMGComp):
            sym_amt = comp.get_el_amt_dict()
            syms = sorted(sym_amt.keys(), key=lambda sym: get_el_sp(sym).X)
            comp = {s: formula_double_format(sym_amt[s], False) for s in syms}

        for e, n in comp.items():
            elems_.append(e)
            nums_.append(n)
        return self.mix_function(elems_, nums_)

    def set_feature_labels(self, values):
        self._feature_labels = [str(s) + "_" + str(n) for s in values for n in range(self.n_composition)]



import torch
import torch.nn as nn

# from bgnet.layer.graph.baselayer import BaseLayer
from featurebox.layer.graph.baselayer import BaseLayer


class AtomLayer(BaseLayer):
    def __init__(self, atom_fea_len, nbr_fea_len, state_fea_len=0):
        """
        Notes:
            N: Total number of atoms in the batch or can be called N_atom.\n
            M: Max number of neighbors (fixed).
        Args:
            atom_fea_len: (int) default 1.
            nbr_fea_len: (int) default 1.
            state_fea_len: (int),default 2.
        """

        super().__init__()

        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.state_fea_len = state_fea_len
        self.fc_full = nn.Linear(2 * self.atom_fea_len + self.nbr_fea_len + self.state_fea_len,
                                 2 * self.atom_fea_len)
        self.fc_full2 = nn.Linear(2 * self.atom_fea_len,
                                  2 * self.atom_fea_len)
        self.sigmoid = nn.Sigmoid()
        self.softplus1 = nn.Softplus()
        self.bn1 = nn.BatchNorm1d(2 * self.atom_fea_len)
        self.bn2 = nn.BatchNorm1d(self.atom_fea_len)
        self.softplus2 = nn.Softplus()

    def forward(self, atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx=None):
        """
        Notes:
            N: Total number of atoms in the batch or can be called N_atom.\n
            M: Max number of neighbors (fixed).\n
            atom_fea_len: default 1.\n
            nbr_fea_len: default 1.\n
            state_fea_len: default 2.
        Args:
            atom_fea: (torch.Tensor) shape (N, atom_fea_len)
                Atom features.
            nbr_fea: (torch.Tensor) shape (N, M, nbr_fea_len)
                Bond features of each atom's M neighbors.
            atom_nbr_idx: (torch.LongTensor) shape (N, M)
                Indices of M neighbors of each atom.
            state_fea: (torch.Tensor) shape (N_node, state_fea_len)
                state_features.
            node_atom_idx: (list of torch.Tensor)
                index for preprocessing.
        Returns:
            atom_out_fea: nn.Variable shape (N, atom_fea_len)
              refreshed atom features.
        """

        N, M = atom_nbr_idx.shape
        # convolution
        atom_nbr_fea = atom_fea[atom_nbr_idx, :]

        if state_fea is not None and self.state_fea_len:
            state_atom_ed = self.expand_idx(state_fea, node_atom_idx)
            total_nbr_fea = torch.cat(
                [atom_fea.unsqueeze(1).expand(N, M, self.atom_fea_len),
                 state_atom_ed.unsqueeze(1).expand(N, M, self.state_fea_len),
                 atom_nbr_fea, nbr_fea], dim=2)
        else:
            total_nbr_fea = torch.cat(
                [atom_fea.unsqueeze(1).expand(N, M, self.atom_fea_len),
                 atom_nbr_fea, nbr_fea], dim=2)

        total_gated_fea = self.fc_full(total_nbr_fea)
        total_gated_fea = self.fc_full2(total_gated_fea)
        total_gated_fea = self.bn1(total_gated_fea.view(
            -1, self.atom_fea_len * 2)).view(N, M, self.atom_fea_len * 2)
        prop_filter, prop_core = total_gated_fea.chunk(2, dim=2)
        prop_filter = self.sigmoid(prop_filter)
        prop_core = self.softplus1(prop_core)
        prop_core = torch.sum(prop_filter * prop_core, dim=1)
        prop_core = self.bn2(prop_core)
        out = self.softplus2(atom_fea + prop_core)
        return out

# if __name__ == "__main__":
#     torch.nn.LayerNorm
#     from bgnet.preprocessing.generator import GraphGenerator, MGEDataLoader
#     from bgnet.layer.graph.test.test_sumlayer import check_get
#
#     check_data = check_get("test/input_data206")
#     data_size = len(check_data[0])
#     gen = GraphGenerator(*check_data, targets=np.arange(data_size))
#     # gen = GraphGenerator(*check_data, targets=None)
#
#     loader = MGEDataLoader(
#         dataset=gen,
#         batch_size=13,
#         shuffle=False,
#         num_workers=0,
#     )
#
#     cl = AtomLayer(16, 1,2)
#
#     loader.to_cuda()
#     cl.to(torch.device("cuda:0"))
#
#     tt.t
#     for time, ((atom_fea,nbr_fea,state_fea,atom_nbr_idx,sgt,oe,node_atom_idx,node_ele_idx,ele_atom_idx),y) in enumerate(loader):
#         # if time ==30:
#         redata = cl(atom_fea, nbr_fea, atom_nbr_idx, state_fea, node_atom_idx)
#         # redata = cl(i[0])
#     tt.t
#     tt.p



import random
from collections import deque
from functools import partial

import numpy as np
from deap import base
from deap import tools
from deap.algorithms import varAnd
from deap.tools import mutShuffleIndexes
from mgetool.newclass import create
from mgetool.tool import batch_parallelize, check_random_state
from sklearn.base import BaseEstimator, MetaEstimatorMixin
from sklearn.datasets import load_boston
from sklearn.feature_selection import SelectorMixin
from sklearn.metrics import r2_score
from sklearn.model_selection import cross_val_score
from sklearn.svm import SVR
from sklearn.utils.validation import check_is_fitted

from featurebox.selection.mutibase import MutiBase


def eaSimple(population, toolbox, cxpb, mutpb, ngen, stats=None, n_jobs=2,
             halloffame=None, verbose=__debug__):
    """This algorithm reproduce the simplest evolutionary algorithm.

    :param population: A list of individuals.
    :param toolbox: A :class:`~deap.base.Toolbox` that contains the evolution
                    operators.
    :param cxpb: The probability of mating two individuals.
    :param mutpb: The probability of mutating an individual.
    :param ngen: The number of generation.
    :param stats: A :class:`~deap.tools.Statistics` object that is updated
                  inplace, optional.
    :param halloffame: A :class:`~deap.tools.HallOfFame` object that will
                       contain the best individuals, optional.
    :param verbose: Whether or not to log the statistics.
    :returns: The final population
    :returns: A class:`~deap.tools.Logbook` with the statistics of the
              evolution

    """
    logbook = tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    # n_job=n
    invalid_ind2 = [tuple(i) for i in invalid_ind]
    fitnesses = batch_parallelize(n_jobs, toolbox.evaluate, invalid_ind2, batch_size=30, tq=True)
    # n_job=1
    # fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if halloffame is not None:
        halloffame.update(population)

    best = deque([halloffame.items[0].fitness.values[0]], maxlen=15)

    # Begin the generational process
    for gen in range(1, ngen + 1):

        # Append the current generation statistics to the logbook
        record = stats.compile(population + halloffame.items) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)
        if verbose:
            print(logbook.stream)

        # Select the next generation individuals
        offspring = toolbox.select(population, len(population) - halloffame.maxsize)
        offspring.extend(halloffame.items)
        # Vary the pool of individuals
        offspring = varAnd(offspring, toolbox, cxpb, mutpb)
        offspring = toolbox.map(toolbox.filt, offspring)
        offspring = list(offspring)
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        # n_job=n
        invalid_ind2 = [tuple(i) for i in invalid_ind]

        fitnesses = batch_parallelize(n_jobs, toolbox.evaluate, invalid_ind2, batch_size=30, tq=True)
        # fitnesses = parallelize(n_jobs, toolbox.evaluate, invalid_ind2)
        # n_job=1
        # fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(offspring)

        best.append(halloffame.items[0].fitness.values[0])
        if sum(best) / best.maxlen == best[-1]:
            break
        # Replace the current population by the offspring
        population[:] = offspring

    return population, logbook


def generate_xi():
    return random.randint(0, 1)


def generate(space):
    return [generate_xi() for _ in range(space)]


def filt(ind, min_=2, max_=None):
    if max_ is not None:
        if np.sum(ind) > max_:
            index = np.where(np.array(ind) == 1)[0]
            k = random.randint(min_, max_)
            index2 = random.choices(index, k=k)
            ind[:] = [0] * len(ind)
            [ind.__setitem__(i, 1) for i in index2]

        elif np.sum(ind) < min_:
            k = random.randint(min_, max_)
            index2 = random.choices(list(range(len(ind))), k=k)
            ind[:] = [0] * len(ind)
            [ind.__setitem__(i, 1) for i in index2]

    else:
        if np.sum(ind) < min_:
            k = random.randint(min_, len(ind))
            index2 = random.choices(list(range(len(ind))), k=k)
            ind[:] = [0] * len(ind)
            [ind.__setitem__(i, 1) for i in index2]

    return ind


class GA(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """
    GA with binding. Please just passing training data.

    Examples
    ---------
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> data = load_boston()
    >>> x = data.data[:50]
    >>> y = data.target[:50]
    >>> svr = SVR(gamma="scale", C=100)
    >>> ga = GA(estimator=svr, n_jobs=2, pop_n=50, hof_n=1, cxpb=0.8, mutpb=0.4, ngen=3, max_or_min="max", mut_indpb=0.1, min_=2, muti_index=[0, 5],random_state=0)
    >>> ga.fit(x_test, y_test)

    Then
    ::

        ga.score_cv(x, y)

    """

    def __init__(self, estimator, n_jobs=2, pop_n=1000, hof_n=1, cxpb=0.6, mutpb=0.3, ngen=40, max_or_min="max",
                 mut_indpb=0.05, max_=None, min_=2, random_state=None, muti_grade=2, muti_index=None):
        """

        Parameters
        ----------
        estimator:
            sklearn estimator
        n_jobs:int
            njobs
        pop_n:int
            population
        hof_n:int
            hof
        cxpb:float
            probility of cross
        mutpb:float
            probility of mutate
        ngen:int
            generation
        max_or_min:str
            "max","min";max problem or min
        mut_indpb:float
            probility of mutate of each node.
        max_:int
            max size
        min_:int
            min size
        random_state:float
            randomstate
        muti_grade:
            binding grade
        muti_index:
            binding range [min,max]
        """
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=None)
        self.estimator = estimator
        self.n_jobs = n_jobs
        self.pop_n = pop_n
        self.hof_n = hof_n
        self.cxpb = cxpb
        self.mutpb = mutpb
        self.ngen = ngen
        self.mut_indpb = mut_indpb
        self.max_ = max_
        self.min_ = min_
        self.max_or_min = max_or_min
        self.random_state = random_state

        check_random_state(random_state)
        random.seed(random_state)
        np.random.seed(0)

        self.toolbox = base.Toolbox()
        if max_or_min == "max":
            FitnessMax = create("FitnessMax", base.Fitness, weights=(1.0,))
        else:
            FitnessMax = create("FitnessMax", base.Fitness, weights=(-1.0,))
        self.Individual = create("Individual", list, fitness=FitnessMax)
        self.toolbox = base.Toolbox()
        # Attribute generator

    def generate_min_max(self, space, min_=2, max_=None):
        ind = [generate_xi() for _ in range(space)]
        return filt(ind, min_=min_, max_=max_)

    def feature_fold_length(self, feature):

        muti_grade, muti_index = self.muti_grade, self.muti_index
        if self.check_muti:
            cc = []
            feature = np.sort(feature)

            i = 0
            while i <= feature[-1]:
                if muti_index[0] <= i < muti_index[1]:
                    i += self.muti_grade
                    cc.append(self.muti_grade)
                else:
                    i += 1
                    cc.append(1)
            return cc
        else:
            return [1] * len(feature)

    def fit(self, X, y):
        """Fit data and run GA."""
        x_space = X.shape[1]
        self.X = X
        self.y = y
        self.x_space_fold = self.feature_fold_length(list(range(x_space)))
        toolbox = self.toolbox
        toolbox.register("generate_x", self.generate_min_max, len(self.x_space_fold), min_=self.min_, max_=self.max_)
        # Structure initializers
        toolbox.register("individual", tools.initIterate, self.Individual,
                         toolbox.generate_x)
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)

        fit_func = partial(self.fitness_func, model=self.estimator, x=X, y=y, return_model=False)
        toolbox.register("evaluate", fit_func)
        toolbox.register("filt", filt, min_=self.min_, max_=self.max_)
        toolbox.register("mate", tools.cxTwoPoint)
        toolbox.register("mutate", mutShuffleIndexes, indpb=self.mut_indpb)
        toolbox.register("select", tools.selTournament, tournsize=3)

        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("min", np.min)
        stats.register("max", np.max)

        pop = toolbox.population(n=self.pop_n)
        self.hof = tools.HallOfFame(self.hof_n)

        eaSimple(pop, self.toolbox, cxpb=self.cxpb, mutpb=self.mutpb, ngen=self.ngen, n_jobs=self.n_jobs,
                 stats=stats, halloffame=self.hof, verbose=True)
        for i in self.hof.items:
            print(self.unfold(i), i.fitness)
        support_ = self.unfold(self.hof.items[0])
        self.support_ = np.array(support_) > 0
        return self

    def unfold(self, ind):
        sss = []
        [sss.extend([i] * j) for i, j in zip(ind, self.x_space_fold)]
        return sss

    def fitness_func(self, ind, model, x, y, return_model=False, cv=5):
        sss = self.unfold(ind)
        index = np.where(np.array(sss) == 1)[0]
        x = x[:, index]
        if x.shape[1] > 1:
            svr = model
            svr.fit(x, y)
            if hasattr(svr, "best_score_"):
                sc = svr.best_score_
            else:
                sc = np.mean(cross_val_score(svr, x, y, cv=cv))
            if return_model:
                return sc, svr
            else:
                return sc,
        else:
            if return_model:
                return 0, None
            else:
                return 0,

    def socre_func(self, ind, model, x, y):
        sss = self.unfold(ind)
        index = np.where(np.array(sss) == 1)[0]
        x = x[:, index]
        if x.shape[1] > 1:
            svr = model
            y2 = svr.predict(x)
            sc = r2_score(y, y2)
            return sc
        else:
            raise TypeError("only one feature,error")

    def predict_func(self, ind, model, x):
        sss = self.unfold(ind)
        index = np.where(np.array(sss) == 1)[0]
        x = x[:, index]
        if x.shape[1] > 1:
            svr = model
            y = svr.predict(x)
            return y
        else:
            raise TypeError("only one feature,error")

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_

    def score_cv(self, X, y):
        """Reduce X to the selected feature and then return the score of the underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """

        mod = self.fitness_func(self.hof.items[0], self.estimator, self.X, self.y, return_model=True)[1]
        score = self.fitness_func(self.hof.items[0], mod, X, y, return_model=False)
        return score

    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """

        mod = self.fitness_func(self.hof.items[0], self.estimator, self.X, self.y, return_model=True)[1]
        score = self.socre_func(self.hof.items[0], mod, X, y)
        return score

    def predict(self, X):
        """Reduce X to the selected feature and then return the score of the underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        """

        mod = self.fitness_func(self.hof.items[0], self.estimator, self.X, self.y, return_model=True)[1]
        score = self.predict_func(self.hof.items[0], mod, X)
        return score


if __name__ == "__main__":
    data = load_boston()
    x = data.data
    y = data.target
    svr = SVR(gamma="scale", C=100)

    ga = GA(estimator=svr, n_jobs=1, pop_n=100, hof_n=1, cxpb=0.8, mutpb=0.4, ngen=10,
            max_or_min="max", mut_indpb=0.1, min_=2, muti_index=[0, 5], random_state=0)
    ga.fit(x, y)
    ga.score(x, y)



"""Get pure atom properties.

Embedded data: "element_table.csv", "elemental_MEGNet.json", "ie.json", "oe.csv"

"""
import functools
import os
from abc import abstractmethod, ABC
from pathlib import Path
from typing import Any, Dict, List, Union, Callable

import numpy as np
import pandas as pd
from monty.serialization import loadfn
from pymatgen.core import Element, Structure, Lattice

from featurebox.data.check_data import ALL_ELE_N_MAP, ALL_N_ELE_MAP
from featurebox.featurizers.base_transform import Converter, BaseFeature

MODULE_DIR = Path(__file__).parent.parent.parent.absolute()


def get_atom_fea_number(structure: Structure) -> List[int]:
    """
    Get atom features from structure, may be overwritten.

    Args:
        structure: (Pymatgen.Structure) pymatgen structure
    Returns:
        List of atomic numbers
    """
    return [i.specie.Z for i in structure]


def get_atom_fea_name(structure: Structure) -> List[dict]:
    """
    For a structure return the list of dictionary for the site occupancy
    for example, Fe0.5Ni0.5 site will be returned as {"Fe": 0.5, "Ni": 0.5}

    Args:
        structure (Structure): pymatgen Structure with potential site disorder

    Returns:
        a list of site fraction description
    """
    return [i.species.as_dict() for i in structure.sites]


##############################################################

class AtomMap(Converter, ABC):
    """
    Base class for atom converter. Map the element type and weight to element data.
    """

    def __init__(self, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'any'):
        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    @staticmethod
    def get_json_embeddings(file_name: str = "elemental_MEGNet.json") -> Dict:
        """get json preprocessing"""
        data = loadfn(MODULE_DIR / "data" / file_name)
        data = {i: np.array(j) for i, j in data.items()}
        return data

    @staticmethod
    def get_csv_embeddings(data_name: str) -> pd.DataFrame:
        """get csv preprocessing"""
        oedata = pd.read_csv(os.path.join(MODULE_DIR, "data", data_name), index_col=0)
        oedata = oedata.apply(pd.to_numeric, errors='ignore')
        oedata = oedata.fillna(0)
        return oedata


class BinaryMap(AtomMap):
    """Base converter with 2 different search_tp."""

    def __init__(self, search_tp: str = "number", weight: bool = True, **kwargs):
        super(BinaryMap, self).__init__(**kwargs)
        self.search_tp = search_tp
        self.weight = weight
        if self.search_tp == "number":
            self.d2 = True
        else:
            self.d2 = False

    def _convert(self, d: Any) -> Any:
        if self.search_tp == "name":
            return self.convert_dict(d)
        else:
            return self.convert_number(d)

    @abstractmethod
    def convert_dict(self, d: List[Dict]):
        """convert list of dict to data"""

        """the subclass must del all the following code and write yourself"""

        # atoms_name = [ALL_ELE_N_MAP[list(i.keys())[0]] for i in d]
        # return self.convert_number(atoms_name)

    @abstractmethod
    def convert_number(self, d: List[int]):
        """convert list of number to data"""
        """The subclass must del all the following code and write yourself"""

        # atoms_name = [{ALL_N_ELE_MAP[i]: 1} for i in d]
        # return self.convert_dict(atoms_name)


# ##############Atom#######################


class AtomJsonMap(BinaryMap):
    """
    Fixed Atom json map.

    Examples
    ----------
    >>> tmps = AtomJsonMap(search_tp="number")
    >>> s = [1,76]                   #[i.specie.Z for i in structure]
    >>> a = tmps.convert(s)
    >>> tmps = AtomJsonMap(search_tp="name")
    >>> s = [{"H": 2, }, {"Al": 1}]  #[i.species.as_dict() for i in pymatgen.structure.sites]
    >>> a = tmps.convert(s)
    >>>
    >>> tmps = AtomJsonMap(search_tp="name")
    >>> s = [[{"H": 2, }, {"Ce": 1}],[{"H": 2, }, {"Al": 1}]]
    >>> a = tmps.transform(s)

    """

    def __init__(self, embedding_dict: Union[str, Dict] = None, search_tp: str = "name", **kwargs):
        """

        Args:
            embedding_dict: (str,dict)
                Name of file or dict,element to element vector dictionary

                Provides the pre-trained elemental embeddings using formation energies,
                which can be used to speed up the training of other models. The embeddings
                are also extremely useful elemental descriptors that encode chemical
                similarity that may be used in other ways.
        """

        super(AtomJsonMap, self).__init__(**kwargs)
        if embedding_dict is None:
            embedding_dict = self.get_json_embeddings()
        elif isinstance(embedding_dict, str):
            embedding_dict = self.get_json_embeddings(embedding_dict)

        assert len(set([len(i) for i in embedding_dict.values()])) == 1, \
            "The element number should be same with `elemental_MEGNet.json`, " \
            "which contains elemental features for 89 elements (up to Z=94, excluding Po, At, Rn, Fr, Ra) "

        self.embedding_dict = embedding_dict
        self.d2 = False
        self.search_tp = search_tp

    def __add__(self, other):
        if isinstance(other, AtomJsonMap):
            for key in other.embedding_dict.keys():
                if key in self.embedding_dict:
                    if isinstance(self.embedding_dict[key], list):
                        self.embedding_dict[key].extend(other.embedding_dict[key])
                    else:
                        self.embedding_dict[key] = np.append(self.embedding_dict[key], other.embedding_dict[key])
                else:
                    pass
        else:
            raise TypeError("only same class can be added.")
        return self

    def convert_dict(self, atoms: List[dict]) -> np.ndarray:
        """
        Convert atom {symbol: fraction} list to numeric features
        """
        if isinstance(atoms, dict):
            atoms = [{k: v} for k, v in atoms.items()]

        features = []
        for atom in atoms:
            emb = 0
            for k, v in atom.items():
                if self.weight:
                    emb += np.array(self.embedding_dict[k]) * v
                else:
                    emb += np.array(self.embedding_dict[k])
            features.append(emb)
        if len(atoms) ==1:
            return np.array(features).ravel()
        else:
            return np.array(features).reshape((len(atoms), -1))

    def convert_number(self, atoms: List[int]) -> np.ndarray:

        atoms_name = [ALL_N_ELE_MAP[i] for i in atoms]
        features = [self.embedding_dict[i] for i in atoms_name]

        if len(atoms) == 1:
            return np.array(features).ravel()
        else:
            return np.array(features).reshape((len(atoms), -1))


class AtomTableMap(BinaryMap):
    """
    Fixed Atom embedding map.
    Default table is oe.csv.
    you can change the table yourself for different preprocessing.
    The table contains elemental features for 92 U elements at least.
    Please check all your data is int or float!!!

    Such as:

    ===== ===== ===== =====
    Data    F0    F1    ...
    ----- ----- ----- -----
    H     V     V     ...
    He    V     V     ...
    Li    V     V     ...
    Be    V     V     ...
    ...   ...   ...   ...
    ===== ===== ===== =====

    Examples
    --------
    >>> tmps = AtomTableMap(search_tp="number")
    >>> s = [1,76]
    >>> a = tmps.convert(s)
    ...
    >>> tmps = AtomTableMap(search_tp="name")
    >>> s = [{"H": 2, }, {"Po": 1}]  #[i.species.as_dict() for i in pymatgen.structure.sites]
    >>> a = tmps.convert(s)
    ...
    >>> tmps = AtomTableMap(search_tp="name",tablename="oe.csv")
    >>> s = [[{"H": 2, }, {"Po": 1}],[{"H": 2, }, {"Po": 1}]]
    >>> a = tmps.transform(s)
    ...

    >>> tmps = AtomTableMap(tablename=None) or
    >>> tmps = AtomTableMap(tablename="element_table.csv")
    >>> s = [{"H": 2, }, {"Pd": 1}]
    >>> b = tmps.convert(s)
    ...

    """

    def __init__(self, tablename: Union[str, np.ndarray, pd.DataFrame] = "oe.csv",
                 search_tp: str = "name", **kwargs):
        """

        Parameters
        ----------
        tablename: str,np.ndarray, pd.Dateframe
            1. Name of table in bgnet.preprocessing.resources. if tablename is None,
            use the embedding "element_table.csv".\n
            2. np.ndarray, search_tp = "number".\n
            3. pd.dataframe, search_tp = "name"

        search_tp:str
            Name
        """

        super(AtomTableMap, self).__init__(search_tp=search_tp, **kwargs)

        if tablename is None or tablename == "element_table.csv":
            self.da = self.get_ele_embeddings()
            self.d2 = True
            self.dax = self.da.values
            self.da_columns = list(self.da.columns)

        elif isinstance(tablename, str):
            self.da = self.get_csv_embeddings(tablename)
            self.dax = self.da.values
            self.da_columns = list(self.da.columns)

        elif isinstance(tablename, np.ndarray):
            self.da_columns = None
            self.dax = tablename
            self.d2 = False
        elif isinstance(tablename, pd.DataFrame):
            self.da = tablename
            self.d2 = True
            self.da_columns = list(self.da.columns)
        else:
            raise TypeError("just accept np.array,pd.dataframe, or str name of csv")
        self.dax = np.concatenate((np.zeros((1, self.dax.shape[1])), self.dax), axis=0)  # for index from 1
        self.tablename = tablename

    @staticmethod
    def get_ele_embeddings() -> pd.DataFrame:
        """get CSV preprocessing"""
        oedata = pd.read_csv(os.path.join(MODULE_DIR, "data", "element_table.csv"), index_col=0, header=7, skiprows=0)
        oedata = oedata.drop(["abbrTex", "abbr"], axis=0)
        oedata = oedata.fillna(0)
        oedata = oedata.apply(pd.to_numeric, errors='ignore')
        return oedata

    def convert_dict(self, atoms: List[Dict]) -> np.ndarray:
        """
        Convert atom {symbol: fraction} list to numeric features
        """
        if isinstance(atoms, dict):
            atoms = [{k: v} for k, v in atoms.items()]

        features = []
        for atom in atoms:
            emb = 0
            for k, v in atom.items():
                if self.weight:
                    emb += self.da.loc[k, :].values * v
                else:
                    emb += self.da.loc[k, :].values
            features.append(emb)
        if len(atoms) == 1:
            return np.array(features).ravel()
        else:
            return np.array(features).reshape((len(atoms), -1))

    def convert_number(self, atoms: List[int]) -> np.ndarray:
        """
        Convert atom number list to numeric features
        """
        return self.dax[atoms, :]

    def __add__(self, other):
        if isinstance(other, AtomTableMap):
            assert other.search_tp == self.search_tp, "should be same"
            if self.search_tp == "number":
                self.dax = np.concatenate((self.dax, other.dax), axis=1)
                self.da = None
            else:
                self.da = pd.concat((self.dax, other.dax), axis=1)
                self.dax = None
        else:
            raise TypeError("only same class can be added.")
        return self

    @property
    def feature_labels(self):
        if self.da_columns is not None:
            return self.da_columns


def process_uni(i):
    """Post-processing for bool preprocessing General properties."""
    dataii = []

    if i is None:
        dataii.append(0.0)

    elif isinstance(i, (list, tuple)):
        if not i:
            dataii.append(0.0)
        else:
            [dataii.extend(process_uni(ii)) for ii in i]

    elif isinstance(i, (int, float)):
        dataii.append(i)
    elif isinstance(i, str):
        dataii.append(i)
    elif isinstance(i, np.ndarray):
        dataii.extend(i.tolist())
    elif isinstance(i, dict):
        if i == {}:
            dataii.append(0.0)
        else:
            dataii.extend(list(i.values()))
    elif isinstance(i, pd.DataFrame):
        dataii.extend(i.values.ravel())
    elif isinstance(i, np.ndarray):
        if i is np.NaN or np.isnan(i):
            dataii.append(0.0)
        else:
            raise TypeError("The {} is cant serialized.".format(i))
    else:
        raise TypeError("The {} is cant serialized.".format(i))

    return dataii


def process_atomic_orbitals(o):
    """Post-processing for dict preprocessing with "1s", "2p", ..."""
    if o is None or {}:
        return [0.0] * 18
    orb = ("1s", "2p", "2s", "3d", "3p", "3s", "4d", "4f", "4p", "4s", "5d", "5f", "5p", "5s", "6d", "6p", "6s", "7s")
    return [o[orbi] if orbi in o else 0.0 for orbi in orb]


def process_tuple_oxidation_states(ox, size=10):
    """Post-processing for tuple of float preprocessing."""
    if ox is None or ():
        return [0.0] * size
    ox = list(ox)
    ox.extend([0.0] * size)
    return ox[:size]


def process_tuple_full_electronic_structure(full_e):
    """Post-processing for electronic_structure preprocessing ( (1,"s",2),...   )"""
    if full_e is None or ():
        return [0.0] * 18
    dic_all = {"{}{}".format(i, j): k for i, j, k in full_e}
    return process_atomic_orbitals(dic_all)


def process_bool_transition_metal(tm):
    """Post-processing for bool preprocessing"""
    if tm is None:
        return [-1, ]
    elif tm is True:
        return [1, ]
    else:
        return [0, ]


after_treatment_func_map_ele = {
    "atomic_orbitals": process_atomic_orbitals,
    "atomic_radius": process_uni,
    "atomic_mass": process_uni,
    "Z": process_uni,
    "X": process_uni,
    "number": process_uni,
    "max_oxidation_state": process_uni,
    "min_oxidation_state": process_uni,
    "oxidation_states": functools.partial(process_tuple_oxidation_states, size=10),
    "common_oxidation_states": functools.partial(process_tuple_oxidation_states, size=10),
    "full_electronic_structure": process_tuple_full_electronic_structure,
    "row": process_uni,
    "group": process_uni,
    "is_transition_metal": process_bool_transition_metal,
    "is_post_transition_metal": process_bool_transition_metal,
    "is_metalloid": process_bool_transition_metal,
    "is_alkali": process_bool_transition_metal,
    "is_alkaline": process_bool_transition_metal,
    "is_halogen": process_bool_transition_metal,
    "is_lanthanoid": process_bool_transition_metal,
    "is_actinoid": process_bool_transition_metal,
    "atomic_radius_calculated": process_uni,
    "van_der_waals_radius": process_uni,
    "mendeleev_no": process_uni,
    "molar_volume": process_uni,
    "thermal_conductivity": process_uni,
    "boiling_point": process_uni,
    "melting_point": process_uni,
    "bulk_modulus": process_uni,
    "youngs_modulus": process_uni,
    "brinell_hardness": process_uni,
    "rigidity_modulus": process_uni,
    "critical_temperature": process_uni,
    "density_of_solid": process_uni,
    "coefficient_of_linear_thermal_expansion": process_uni,
    "average_ionic_radius": process_uni,
    "average_cationic_radius": process_uni,
    "average_anionic_radius": process_uni,
    "ionic_radii": functools.partial(process_tuple_oxidation_states, size=6),
}


class AtomPymatgenPropMap(BinaryMap):
    """
    Get pymatgen element preprocessing.
    prop_name = [
    "atomic_radius",
    "atomic_mass",
    "number",
    "max_oxidation_state",
    "min_oxidation_state",
    "row",
    "group",
    "atomic_radius_calculated",
    "mendeleev_no",
    "critical_temperature",
    "density_of_solid",
    "average_ionic_radius",
    "average_cationic_radius",
    "average_anionic_radius",]

    Examples
    --------
    >>> tmps = AtomPymatgenPropMap(search_tp="number",prop_name=["X"])
    >>> s = [1,76]
    >>> a = tmps.convert(s)
    >>> tmps = AtomPymatgenPropMap(search_tp="name",prop_name=["X"])
    >>> s = [{"H": 2, }, {"Po": 1}]  #[i.species.as_dict() for i in pymatgen.structure.sites]
    >>> a = tmps.convert(s)
    >>> tmps = AtomPymatgenPropMap(search_tp="name",prop_name=["X"])
    >>> s = [[{"H": 2, }, {"Po": 1}],[{"H": 2, }, {"Po": 1}]]
    >>> a = tmps.transform(s)

    """

    def __init__(self, prop_name: Union[str, List[str]], func: Callable = None, search_tp: str = "name", **kwargs):
        """

        Args:
            prop_name:(str,list of str)
                prop name or list of prop name
            func:(callable or list of callable)
                please make sure the size of it is the same with prop_name.
            search_tp:(str)
                location method.
                "name" for dict
                "number" for int.
        """
        super(AtomPymatgenPropMap, self).__init__(search_tp=search_tp, **kwargs)

        if isinstance(prop_name, (list, tuple)):
            self.prop_name = list(prop_name)
        else:
            self.prop_name = [prop_name, ]

        if func is None:
            func = len(self.prop_name) * [process_uni]
        if isinstance(func, (list, tuple)):
            self.func = list(func)
        else:
            self.func = [func, ]

        if len(self.func) == 1 and len(self.prop_name) > 1:
            self.func *= len(self.prop_name)

        assert len(self.prop_name) == len(self.func), "The size of prop and func should be same."
        self.func = [process_uni if i is None else i for i in self.func]

        for i, j in enumerate(self.prop_name):
            if j in after_treatment_func_map_ele:
                self.func[i] = after_treatment_func_map_ele[j]
        self.da = [Element.from_Z(i) for i in range(1, 119)]
        self.da.insert(0, None) #for start from 1
        self.ele_map = []
        self.d2 = False

    def convert_dict(self, atoms: List[Dict]) -> np.ndarray:
        """
        Convert atom {symbol: fraction} list to numeric features
        """
        if isinstance(atoms, dict):
            atoms = [{k: v} for k, v in atoms.items()]

        features = []
        for atom in atoms:
            emb = 0
            for k, v in atom.items():
                k = ALL_ELE_N_MAP[k]
                data_all = []
                datai = [getattr(self.da[k], pi) for pi in self.prop_name]
                datai = [self.func[i](j) for i, j in enumerate(datai)]
                [data_all.extend(i) for i in datai]
                try:
                    if self.weight:
                        emb += np.array(data_all) * v
                    else:
                        emb += np.array(data_all)
                except BaseException:
                    raise TypeError("try add {} and {}".format(emb, datai),
                                    "with dtype {} and {}".format(np.array(emb).dtype, np.array(datai).dtype),
                                    "with size {} and {}".format(np.array(emb).shape, np.array(datai).shape),
                                    "The preprocessing cannot be convert in to float, "
                                    "or the preprocessing after func are with different size.",
                                    "please check you func, which keep the number of results consistent.")

            features.append(emb)
        if len(set([i.shape for i in features])) != 1:
            print("The preprocessing after func are with different size.",
                  "please check you func, which keep the number of results consistent.")
        if len(atoms) == 1:
            return np.array(features).ravel()
        else:
            return np.array(features).reshape((len(atoms), -1))

    def convert_number(self, atoms: List[int]) -> np.ndarray:
        """
        Convert int list to numeric features
        """
        features = []
        for atom in atoms:
            data_all = []
            datai = [getattr(self.da[round(atom)], pi) for pi in self.prop_name]
            datai = [self.func[i](j) for i, j in enumerate(datai)]
            [data_all.extend(i) for i in datai]
            emb = np.array(data_all)
            features.append(emb)
        if len(set([i.shape for i in features])) != 1:
            print("The preprocessing after func are with different size.",
                  "please check you func, which keep the number of results consistent.")
        if len(atoms) == 1:
            return np.array(features).ravel()
        else:
            return np.array(features).reshape((len(atoms), -1))

    def __add__(self, other):
        if isinstance(other, AtomPymatgenPropMap):
            assert other.search_tp == self.search_tp, "should be same"
            self.prop_name.extend(other.prop_name)
            self.func.extend(other.func)
        else:
            raise TypeError("only same class can be added.")
        return self

    @property
    def feature_labels(self):
        return self.prop_name


after_treatment_func_map_structure = {
    "atomic_orbitals": process_atomic_orbitals,
    "density": process_uni,
}


def _getter_arr(obj, pi):
    """Get prop.
    """
    if "." in pi:
        pis = list(pi.split("."))
        pis.reverse()
        while len(pis):
            s = pis.pop()
            obj = _getter_arr(obj, s)
        return obj
    elif "()" in pi:
        return getattr(obj, pi[:-2])()

    else:
        return getattr(obj, pi)


class _StructurePymatgenPropMap(BaseFeature):
    """
    Get property of pymatgen structure preprocessing.

    Examples
    -----------
    tmps = StructurePymatgenPropMap()

    """

    def __init__(self, prop_name=None, func: Callable = None, return_type="df", **kwargs):
        """

        Args:
            prop_name:(str,list of str)
                prop name or list of prop name
                default ["density", "volume", "ntypesp"]
            func:(callable or list of callable)
                please make sure the size of it is the same with prop_name.
        """
        super(_StructurePymatgenPropMap, self).__init__(return_type=return_type, **kwargs)

        if prop_name is None:
            prop_name = ["density", "volume", "ntypesp"]
        if isinstance(prop_name, (list, tuple)):
            self.prop_name = list(prop_name)
        else:
            self.prop_name = [prop_name, ]

        if func is None:
            func = len(self.prop_name) * [process_uni]
        if isinstance(func, (list, tuple)):
            self.func = list(func)
        else:
            self.func = [func, ]

        if len(self.func) == 1 and len(self.prop_name) > 1:
            self.func *= len(self.prop_name)

        assert len(self.prop_name) == len(self.func), "The size of prop and func should be same."
        self.func = [process_uni if i is None else i for i in self.func]

        for i, j in enumerate(self.prop_name):
            if j in after_treatment_func_map_structure:
                self.func[i] = after_treatment_func_map_structure[j]
        self.d2 = False
        self.lengths = []

    def convert(self, structure: [Structure, Lattice]) -> np.ndarray:
        data_all = []
        datai = []
        for pi in self.prop_name:
            datai.append(_getter_arr(structure, pi))
        datai = [self.func[i](j) for i, j in enumerate(datai)]
        if not self.lengths:
            self.lengths = [len(i) if isinstance(i, (np.ndarray, tuple, list)) else 1 for i in datai]
        [data_all.extend(i) for i in datai]

        return np.array(data_all).ravel()

    def __add__(self, other):
        if isinstance(other, AtomPymatgenPropMap):
            self.prop_name.extend(other.prop_name)
            self.func.extend(other.func)
        else:
            raise TypeError("only same class can be added.")
        return self

    @property
    def feature_labels(self):
        nall = []
        for p, l in zip(self.prop_name, self.lengths):
            if l == 1:
                ni = [p, ]
            else:
                ni = []
                for i in range(l):
                    ni.append("_".join([str(p), str(i)]))

            nall.extend(ni)
        self._feature_labels = nall
        return self._feature_labels



"""Use descriptors form ``pyXtal_FF``, in :mod:`featurebox.featurizers.descriptors`
"""
from featurebox.featurizers.descriptors.ACSF import ACSF
from featurebox.featurizers.descriptors.EAD import EAD
from featurebox.featurizers.descriptors.EAMD import EAMD
from featurebox.featurizers.descriptors.SO3 import SO3
from featurebox.featurizers.descriptors.SO4 import SO4_Bispectrum
from featurebox.featurizers.descriptors.SOAP import SOAP
from featurebox.featurizers.descriptors.behlerparrinello import BehlerParrinello
from featurebox.featurizers.descriptors.wACSF import wACSF
from featurebox.utils.look_json import mark_classes

DesDict = mark_classes([ACSF,
                        BehlerParrinello,
                        EAD,
                        EAMD,
                        SOAP,
                        SO3,
                        SO4_Bispectrum,
                        wACSF,
                        ])

for i, j in DesDict.items():
    locals()[i] = j



# -*- coding: utf-8 -*-

# @Time    : 2020/11/27 11:27
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause

from itertools import zip_longest
from typing import List, Dict

import pandas as pd
from pymatgen.core import Structure
from pymatgen.ext.matproj import MPRester
from tqdm import tqdm


class MpAccess:
    """
    API for pymatgen database, access pymatgen to get data.

    Examples
    --------
    >>> mpa = MpAccess("Di28ZMunseR8vr57")
    >>> ids = mpa.get_ids({"elements": {"$in": ["Al","O"]},'nelements': {"$lt": 2, "$gte": 1}})
    number 29
    >>> df = mpa.data_fetcher(mp_ids=ids, mp_props=['material_id', "cif"])
    Will fetch 29 inorganic compounds from Materials Project
    >>> structures_list = mpa.cifs_to_structures()
    ...

    """

    def __init__(self, api_key: str = "Di28ZMunseR8vr56"):
        """

        Parameters
        ----------
        api_key:str:
            pymatgen key.
        """
        self.m = MPRester(api_key)
        self.dff = None
        self.ids = None

    def data_fetcher(self, mp_ids: List[str] = None, mp_props: List[str] = None,
                     elasticity: bool = False) -> pd.DataFrame:
        """
        Fetch file from pymatgen.

        prop_name=['band_gap','density',"icsd_ids"'volume','material_id','pretty_formula','elements',"energy",
        'efermi','e_above_hull','formation_energy_per_atom','final_energy_per_atom','unit_cell_formula',
        'spacegroup','nelements'"nsites","final_structure","cif","piezo","diel"]

        Parameters
        ----------
        mp_ids:list of str
            list of MP id of pymatgen.
        mp_props:list of str
            prop_names
        elasticity:bool
            obtain elasticity or not.

        Returns
        -------
        pandas.DataFrame
            properties Table.
        """

        print('Will fetch %s inorganic compounds from Materials Project' % len(mp_ids))

        def grouper(iterable, n, fillvalue=None):
            """"
            Split requests into fixed number groups
            eg: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
            Collect data_cluster into fixed-length chunks or blocks"""
            args = [iter(iterable)] * n
            return zip_longest(fillvalue=fillvalue, *args)

        if mp_ids is None:
            mp_ids = self.ids
        # the following props will be fetched
        if mp_props is None:
            mp_props = [
                'band_gap',
                'density',
                "icsd_ids"
                'volume',
                'material_id',
                'pretty_formula',
                'elements',
                "energy",
                'efermi',
                'e_above_hull',
                'formation_energy_per_atom',
                'final_energy_per_atom',
                'unit_cell_formula',
                'spacegroup',
                'nelements'
                "nsites",
                "final_structure",
                "cif",
                "piezo",
                "diel"
            ]

        if elasticity:
            mp_props.append("elasticity")

        entries = []
        mpid_groups = [g for g in grouper(mp_ids, 40)]

        with self.m as mpr:
            for group in tqdm(mpid_groups):
                mpid_list = [ids for ids in filter(None, group)]
                chunk = mpr.query({"material_id": {"$in": mpid_list}}, mp_props)
                entries.extend(chunk)

        if 'material_id' in entries[0]:
            df = pd.DataFrame(entries, index=[e['material_id'] for e in entries])
        else:
            df = pd.DataFrame(entries)
        if 'unit_cell_formula' in df.columns:
            df = df.rename(columns={'unit_cell_formula': 'composition'})

        df = df.T

        self.dff = df

        return df

    def cifs_to_structures(self, cifs: List[str] = None) -> List[Structure]:
        """Get structures from cifs"""
        if cifs is None:
            if "cif" in self.dff.index:
                cifs = self.dff.loc["cif"]
            else:
                raise TypeError("cif is not in data")

        return [Structure.from_str(i, "cif") for i in cifs]

    def get_ids(self, criteria: Dict = None):
        """
        Search id by criteria.

        support_property = ['energy', 'energy_per_atom', 'volume', 'formation_energy_per_atom', 'nsites',
        'unit_cell_formula','pretty_formula', 'is_hubbard', 'elements', 'nelements', 'e_above_hull', 'hubbards',
        'is_compatible', 'spacegroup', 'task_ids',  'band_gap', 'density', 'icsd_id', 'icsd_ids', 'cif',
        'total_magnetization','material_id', 'oxide_type', 'tags', 'elasticity']

        Examples
        --------
        >>> from itertools import combinations
        >>> name_list = ["NaCl","CaCo3"]
        >>> criteria = {
        ... 'pretty_formula': {"$in": name_list},
        ... 'nelements': {"$lt": 3, "$gte": 3},
        ... 'spacegroup.number': {"$in": [225]},
        ... 'crystal_system': "cubic",
        ... 'nsites': {"$lt": 20},
        ... 'formation_energy_per_atom': {"$lt": 0},
        ... "elements": {"$all": "O"},
        ... "piezo":{"$ne": None}
        ... "elements": {"$all": "O"},
        ... "elements": {"$in": list(combinations(["Al", "Co", "Cr", "Cu", "Fe", 'Ni'], 5))}}

        where, ``"$gt"`` >,  ``"$gte"`` >=,  ``"$lt"`` <,  ``"$lte"`` <=,  ``"$ne"`` !=,  ``"$in"``,  ``"$nin"`` (not in),
        ``"$or"``,  ``"$and"``,  ``"$not"``,  ``"$nor"`` ,  ``"$all"``
        """

        if criteria is None:
            criteria = {}

        ids = self.m.query(
            criteria=criteria,
            properties=["material_id"])
        print("number %s" % len(ids))
        ids = pd.DataFrame(ids).values.ravel().tolist()
        self.ids = ids
        return ids

    def get_ids_from_web_table(self, path_file: str = None) -> List[str]:
        """This is a add method to read csv file download from web,the file name is '_Materials Project.csv',
        which contains "Materials Id" columns.
        """
        if path_file is None:
            path_file = "_Materials Project.csv"
        ids_data = pd.read_csv(path_file)
        ids = ids_data["Materials Id"]
        self.ids = ids
        return ids


if __name__ == "__main__":
    #############

    mpa = MpAccess("Di28ZMunseR8vr56")
    ids = mpa.get_ids({"elements": {"$in": ["Al", "O"]}, 'nelements': {"$lt": 2, "$gte": 1}})

    df = mpa.data_fetcher(mp_ids=ids, mp_props=['material_id', "cif"])
    structures_list = mpa.cifs_to_structures()



import torch.nn as nn

# from bgnet.layer.graph.atomlayer import AtomLayer
# from bgnet.layer.graph.baselayer import BaseLayer
# from bgnet.layer.graph.bondlayer import BondLayer
# from bgnet.layer.graph.statelayer import StateLayer
from featurebox.layer.graph.atomlayer import AtomLayer
from featurebox.layer.graph.baselayer import BaseLayer
from featurebox.layer.graph.bondlayer import BondLayer
from featurebox.layer.graph.statelayer import StateLayer


class Block(nn.Module):
    def __init__(self, atom_fea_len, nbr_fea_len, state_fea_len=0):
        super().__init__()
        self.atomlayer = AtomLayer(atom_fea_len, nbr_fea_len, state_fea_len)
        self.bondlayer = BondLayer(atom_fea_len, nbr_fea_len, state_fea_len)
        self.statelayer = StateLayer(atom_fea_len, nbr_fea_len, state_fea_len)

    def forward(self, atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx):
        atom_fea = self.atomlayer(atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx)
        nbr_fea = self.bondlayer(atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx)
        state_fea = self.statelayer(atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx)
        return atom_fea, nbr_fea, state_fea


class MEGNet(BaseLayer):
    def __init__(self, atom_fea_len, nbr_fea_len, state_fea_len=2,
                 inner_atom_fea_len=64, n_conv=2, h_fea_len=128, n_h=1,
                 classification=False, class_number=2):
        """

        Parameters
        ----------
        inner_atom_fea_len: int
          Number of atom features in the input.
        nbr_fea_len: int
          Number of bond features.
        inner_atom_fea_len: int
          Number of hidden atom features in the convolutional layers
        n_conv: int
          Number of convolutional layers
        h_fea_len: int,tuple
          Number of hidden features after merge to samples
        n_h: int
          Number of hidden layers after merge to samples
        """
        super().__init__()
        self.classification = classification
        self.mes = ("mean", "sum", "min", "max")
        le = len(self.mes)
        if isinstance(h_fea_len, int):
            h_fea_len = tuple([h_fea_len for _ in range(n_h)])

        # change feature length
        self.embedding = nn.Linear(atom_fea_len, inner_atom_fea_len)
        # conv
        self.convs = nn.ModuleList([AtomLayer(atom_fea_len=inner_atom_fea_len,
                                              nbr_fea_len=nbr_fea_len)
                                    for _ in range(n_conv)])

        # self.merge_idx_methods expand self.mes times

        # conv to linear
        self.conv_to_fc = nn.Linear(le * inner_atom_fea_len, h_fea_len[0])
        self.conv_to_fc_softplus = nn.Softplus()
        n_h = len(h_fea_len)
        # linear (connect)
        if self.classification:
            self.dropout = nn.Dropout()
        if n_h >= 2:
            self.fcs = nn.ModuleList([nn.Linear(h_fea_len[hi], h_fea_len[hi + 1])
                                      for hi in range(n_h - 1)])
            self.softpluses = nn.ModuleList([nn.Softplus()
                                             for _ in range(n_h - 1)])
        # linear out
        if self.classification:
            self.fc_out = nn.Linear(h_fea_len, class_number)
            self.logsoftmax = nn.LogSoftmax(dim=1)
        else:
            self.fc_out = nn.Linear(h_fea_len[-1], 1)

    def forward(self, atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx, *args, **kwargs):
        """
        Forward pass
        N: Total number of atoms in the batch.\n
        M: Max number of neighbors.\n
        N0: Total number of crystals in the batch

        Parameters
        ----------
        atom_fea: Variable(torch.Tensor) shape (N, orig_atom_fea_len)
            Atom features from atom type
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
            Bond features of each atom's M neighbors
        state_fea: (torch.Tensor) torch.float32, shape (N_node, state_fea_len)
            state feature.
        atom_nbr_idx: torch.LongTensor shape (N, M)
            Indices of M neighbors of each atom
        node_atom_idx: list of torch.LongTensor of length N0
            Mapping from the crystal idx to atom idx

        Returns
        -------
        prediction: nn.Variable shape (N, )
            Atom hidden features after convolution
        """
        atom_fea = self.embedding(atom_fea)
        for conv_func in self.convs:
            atom_fea = conv_func(atom_fea, nbr_fea, state_fea=state_fea, atom_nbr_idx=atom_nbr_idx)
        crys_fea = self.merge_idx_methods(atom_fea, node_atom_idx, methods=self.mes)
        crys_fea = self.conv_to_fc(crys_fea)
        crys_fea = self.conv_to_fc_softplus(crys_fea)

        if self.classification:
            crys_fea = self.dropout(crys_fea)
        if hasattr(self, 'fcs') and hasattr(self, 'softpluses'):
            for fc, softplus in zip(self.fcs, self.softpluses):
                crys_fea = softplus(fc(crys_fea))
        out = self.fc_out(crys_fea)

        if self.classification:
            out = self.logsoftmax(out)
        return out



import mpmath
import numpy as np

lmax = 32

mpmath.mp.dps = 4 * lmax

_Wigner_coefficients = np.empty(((8 * lmax ** 3 + 18 * lmax ** 2 + 13 * lmax + 3) // 3,), dtype=float)
i = 0
for twol in range(0, 2 * lmax + 1):
    for twomp in range(-twol, twol + 1, 2):
        for twom in range(-twol, twol + 1, 2):
            tworho_min = max(0, twomp - twom)
            _Wigner_coefficients[i] = float(mpmath.sqrt(mpmath.fac((twol + twom) // 2) * mpmath.fac((twol - twom) // 2)
                                                        / (mpmath.fac((twol + twomp) // 2) * mpmath.fac(
                (twol - twomp) // 2)))
                                            * mpmath.fac((twol + twomp) // 2)
                                            * mpmath.fac((twol - twomp) // 2))
            i += 1
print(i, _Wigner_coefficients.shape)
np.save('Wigner_coefficients', _Wigner_coefficients)



import unittest

import numpy as np
import pandas as pd

from featurebox.featurizers.base_graph import CrystalGraph, CrystalGraphWithBondTypes, CrystalGraphDisordered


class TestCrystalGraph(unittest.TestCase):
    def setUp(self) -> None:
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]

    def test_CrystalGraph_convert_call(self):
        sg1 = CrystalGraph()
        s11 = sg1(self.data0)
        s12 = sg1(self.data0, state_attributes=np.array([2, 3.0]))

        self.assertTrue(isinstance(s12, dict))
        self.assertEqual(list(s12.keys()), ['atom', 'bond', 'state', 'atom_nbr_idx'])
        for i in s12.values():
            print(type(i))
            self.assertTrue(isinstance(i, (np.ndarray, list)))

    def test_CrystalGraph_as_dict(self):
        sg1 = CrystalGraph()
        dict1 = sg1.as_dict()
        sg2 = CrystalGraph.from_dict(dict1)
        s12 = sg2(self.data0)


class TestCrystalGraphWithBondTypes(unittest.TestCase):
    def setUp(self) -> None:
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]

    def test_CrystalGraph_convert_call(self):
        sg1 = CrystalGraphWithBondTypes()
        s11 = sg1(self.data0)
        s12 = sg1(self.data0, state_attributes=np.array([2, 3.0]))

        self.assertTrue(isinstance(s12, dict))
        self.assertEqual(list(s12.keys()), ['atom', 'bond', 'state', 'atom_nbr_idx'])
        for i in s12.values():
            print(type(i))
            self.assertTrue(isinstance(i, (np.ndarray, list)))

    def test_CrystalGraph_as_dict(self):
        sg1 = CrystalGraphWithBondTypes()
        dict1 = sg1.as_dict()
        sg2 = CrystalGraphWithBondTypes.from_dict(dict1)
        s12 = sg2(self.data0)


class TestCrystalGraphDisordered(unittest.TestCase):
    def setUp(self) -> None:
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]

    def test_CrystalGraph_convert_call(self):
        sg1 = CrystalGraphDisordered()
        s11 = sg1(self.data0)
        s12 = sg1(self.data0, state_attributes=np.array([2, 3.0]))

        self.assertTrue(isinstance(s12, dict))
        self.assertEqual(list(s12.keys()), ['atom', 'bond', 'state', 'atom_nbr_idx'])
        for i in s12.values():
            print(type(i))
            self.assertTrue(isinstance(i, (np.ndarray, list)))

    def test_CrystalGraph_as_dict(self):
        sg1 = CrystalGraphDisordered()
        dict1 = sg1.as_dict()
        sg2 = CrystalGraphDisordered.from_dict(dict1)
        s12 = sg2(self.data0)


if __name__ == '__main__':
    unittest.main()



from collections import Callable

from featurebox.featurizers.atom.mapper import _StructurePymatgenPropMap


# re-site for the nature of the classification.
class StructurePymatgenPropMap(_StructurePymatgenPropMap):
    """
    Get property of pymatgen structure preprocessing.
    default ["density", "volume", "ntypesp"]

    Examples
    -----------
    >>> tmps = StructurePymatgenPropMap()
    >>> tmps.fit_transform()

    """

    def __init__(self, prop_name=None, func: Callable = None, return_type="df", **kwargs):
        """
        Args:
            prop_name:(str,list of str)
                prop name or list of prop name
                default ["density", "volume", "ntypesp"]
            func:(callable or list of callable)
                please make sure the size of it is the same with prop_name.
        """
        super(StructurePymatgenPropMap, self).__init__(prop_name, func, return_type, **kwargs)



"""Use NearNeighbors from ``pymatgen``.

Most in :mod:`pymatgen.analysis.local_env` or
in :func:`pymatgen.optimization.neighbors.find_points_in_spheres`
The costumed as following:
"""
from typing import Dict, List

from pymatgen.core import Element
from pymatgen.analysis.local_env import (
    NearNeighbors,
    VoronoiNN,
    MinimumDistanceNN,
    MinimumVIRENN,
    BrunnerNN_reciprocal,
    BrunnerNN_real,
    BrunnerNN_relative,
    EconNN,
    CrystalNN,
)
from pymatgen.core import Molecule
from pymatgen.core.structure import Structure

from featurebox.utils.look_json import mark_classes


def _is_in_targets(site, targets):
    """
    Test whether a site contains elements in the target list

    Args:
        site (Site): Site to assess
        targets ([Element]) List of elements
    Returns:
         (boolean) Whether this site contains a certain list of elements
    """
    elems = _get_elements(site)
    for elem in elems:
        if elem not in targets:
            return False
    return True


def _get_elements(site):
    """
    Get the list of elements for a Site

    Args: site (Site): Site to assess
    Returns: [Element]: List of elements
    """
    try:
        if isinstance(site.specie, Element):
            return [site.specie]
        return [Element(site.specie)]
    except Exception:
        return site.species.elements


class MinimumDistanceNNAll(NearNeighbors):
    """
    Determine bonded sites by fixed cutoff.
    """

    def __init__(self, cutoff: float = 4.0):
        """
        Args:
            cutoff (float): cutoff radius in Angstrom to look for trial
                near-neighbor sites (default: 4.0).
        """
        self.cutoff = cutoff

    def get_nn_info(self, structure: Structure, n: int) -> List[Dict]:
        """
        Get all near-neighbor sites as well as the associated image locations
        and weights of the site with index n using the closest neighbor
        distance-based method.

        Args:
            structure (Structure): input structure.
            n (int): index of site for which to determine near
                neighbors.

        Returns:
            (list of tuples (Site, array, float)):
            tuples, each one of which represents a neighbor site, its image location,
                and its weight.
        """

        site = structure[n]
        neighs_dists = structure.get_neighbors(site, self.cutoff)

        siw = []
        for nn in neighs_dists:
            siw.append(
                {
                    "site": nn,
                    "image": self._get_image(structure, nn),
                    "weight": nn.nn_distance,
                    "site_index": self._get_original_site(structure, nn),
                }
            )
        return siw


class AllAtomPairs(NearNeighbors):
    """
    Get all combinations of atoms as bonds in a molecule
    """

    def get_nn_info(self, molecule: Molecule, n: int) -> List[Dict]:
        """
        Get near neighbor information
        Args:
            molecule (Molecule): pymatgen Molecule
            n (int): number of molecule

        Returns: List of neighbor dictionary

        """
        site = molecule[n]
        siw = []
        for i, s in enumerate(molecule):
            if i != n:
                siw.append({"site": s, "image": None, "weight": site.distance(s), "site_index": i})
        return siw


class UserVoronoiNN(VoronoiNN):

    def _extract_nn_info(self, structure, nns):
        """Given Voronoi NNs, extract the NN info in the form needed by NearestNeighbors

        Args:
            structure (Structure): Structure being evaluated
            nns ([dicts]): Nearest neighbor information for a structure
        Returns:
            (list of tuples (Site, array, float)): See nn_info
        """

        # Get the target information
        if self.targets is None:
            targets = structure.composition.elements
        else:
            targets = self.targets

        # Extract the NN info
        siw = []
        max_weight = max(nn[self.weight] for nn in nns.values())
        for nstats in nns.values():
            site = nstats["site"]
            if nstats[self.weight] > self.tol * max_weight and _is_in_targets(
                    site, targets
            ):
                nn_info = {
                    "site": site,
                    "image": self._get_image(structure, site),
                    "weight": nstats[self.weight] / max_weight,
                    "site_index": self._get_original_site(structure, site),
                }

                if self.extra_nn_info:
                    # Add all the information about the site
                    poly_info = nstats
                    del poly_info["site"]
                    nn_info["poly_info"] = poly_info
                    nn_info["add_image"] = (poly_info["solid_angle"], poly_info["volume"], poly_info["face_dist"],
                                            poly_info["area"], poly_info["n_verts"])
                siw.append(nn_info)
        return siw


NNDict = mark_classes([
    VoronoiNN,
    UserVoronoiNN,
    MinimumDistanceNN,
    MinimumDistanceNNAll,
    AllAtomPairs,
    # OpenBabelNN,  #BabelMolAdaptor requires openbabel to be installed with Python bindings.
    # CovalentBondNN, just for molecule
    MinimumVIRENN,
    # MinimumOKeeffeNN, # just for specific element/
    BrunnerNN_reciprocal,
    BrunnerNN_real,
    BrunnerNN_relative,
    EconNN,
    CrystalNN,
    # CutOffDictNN, # just for specific element/
    # Critic2NN, #Critic2Caller requires the executable critic to be in the path.
    # Please follow the instructions at https://github.com/aoterodelaroza/critic2.
])

for i, j in NNDict.items():
    locals()[i] = j



# -*- coding: utf-8 -*-

# @Time   : 2019/5/25 17:28
# @Author : Administrator
# @Project : feature_preparation
# @FileName: exhaustion.py
# @Software: PyCharm

import warnings
from functools import partial
from itertools import combinations
from typing import Tuple, List

import numpy as np
from mgetool.tool import parallelize
from sklearn.base import BaseEstimator
from sklearn.base import MetaEstimatorMixin
from sklearn.base import clone
from sklearn.feature_selection import SelectorMixin
from sklearn.model_selection import cross_val_score
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils.validation import check_is_fitted, check_X_y

from featurebox.selection.mutibase import MutiBase

warnings.filterwarnings("ignore")


class Exhaustion(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """
    Exhaustion features combination.

    The attribute ``estimator_`` is the model with the best feature rather than all feature combination.

    Examples
    ----------
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> bf = Exhaustion(svr,n_select=(2,),refit=True)
    >>> new_x = bf.fit_transform(X,y)
    >>> bf.support_
    array([False, False, False, False, False, False, False, False, False,
           False,  True, False,  True])

    Examples
    ----------
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> from sklearn import model_selection
    >>> gd = model_selection.GridSearchCV(svr, param_grid=[{"C": [1, 10]}], n_jobs=1, cv=5)
    >>> bf = Exhaustion(gd,n_select=(2,),refit=True)
    >>> new_x = bf.fit_transform(X,y)
    >>> bf.support_
    ...
    """

    def __init__(self, estimator: BaseEstimator, n_select: Tuple = (2, 3, 4), muti_grade: int = 2,
                 muti_index: List = None, must_index: List = None, n_jobs: int = 1,
                 refit: bool = False, cv: int = 5):
        """

        Parameters
        ----------
        estimator:
            sklearn model or GridSearchCV
        n_select:tuple
            the n_select list,default,n_select=(3, 4)
        muti_grade:list
            binding_group size, calculate the correction between binding
        muti_index:list
            the range of muti_grade:[min,max)
        must_index:list
            the columns force to index
        n_jobs:int
            n_jobs
        refit:bool
            refit or not, if refit the model would used all data.
        cv:bool
            if estimator is sklearn model, used cv, else pass
        """
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.estimator = estimator
        self.score_ = []
        self.n_jobs = n_jobs
        self.n_select = [n_select, ] if isinstance(n_select, int) else n_select
        self.refit = refit
        self.cv = cv

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    def fit(self, X, y):
        """Fit the baf model and then the underlying estimator on the selected feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            The training input0 samples.

        y : array-like, shape = [n_samples]
            The target values.
        """
        return self._fit(X, y)

    def _fit(self, x, y):

        estimator = clone(self.estimator)

        def score_pri(slices, x0, y0):
            slices = list(slices)
            if len(slices) < 1:
                score0 = - np.inf
            else:
                slices = self.feature_unfold(slices)
                data_x0 = x0[:, slices]

                if hasattr(estimator, "best_score_"):
                    estimator.fit(data_x0, y0)
                    score0 = np.mean(estimator.best_score_)  # score_test

                else:
                    score0 = cross_val_score(estimator, data_x0, y0, cv=self.cv)
                    score0 = np.mean(score0)
                # print(slices, score0)
            return score0

        score = partial(score_pri, x0=x, y0=y)

        self.score_ = []
        x, y = check_X_y(x, y, "csc")
        assert all((self.check_must, self.check_muti)) in [True, False]

        feature_list = list(range(x.shape[1]))
        fold_feature_list = self.feature_fold(feature_list)
        if self.check_must:
            fold_feature_list = [i for i in fold_feature_list if i not in self.check_must]

        slice_all = [combinations(fold_feature_list, i) for i in self.n_select]
        slice_all = [list(self.feature_must_fold(_)) for i in slice_all for _ in i]

        scores = parallelize(n_jobs=self.n_jobs, func=score, iterable=slice_all)

        feature_combination = [self.feature_unfold(_) for _ in slice_all]
        index = np.argmax(scores)
        select_feature = feature_combination[int(index)]
        su = np.zeros(x.shape[1], dtype=np.bool)
        su[select_feature] = 1
        self.best_score_ = max(scores)
        self.score_ = scores
        self.support_ = su
        self.estimator_ = clone(self.estimator)
        if self.refit:
            if not hasattr(self.estimator_, 'best_score_'):
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            if hasattr(self.estimator_, 'best_score_') and hasattr(self.estimator_, "refit") \
                    and self.estimator_.refit is True:
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            self.estimator_.fit(x[:, select_feature], y)
        self.n_feature_ = len(select_feature)
        self.score_ex = list(zip(feature_combination, scores))
        self.scatter = list(zip([len(i) for i in slice_all], scores))
        self.score_ex.sort(key=lambda _: _[1], reverse=True)

        return self

    @if_delegate_has_method(delegate='estimator_')
    def predict(self, X):
        """Reduce X to the selected feature and then Fit using the underlying estimator.
        Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator_')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the underlying estimator.
        Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_


ExhaustionCV = Exhaustion



import unittest

import pandas as pd

from featurebox.data.check_data import CheckElements
from featurebox.featurizers.base_graph import CrystalGraph
from featurebox.featurizers.generator import GraphGenerator, MGEDataLoader


class Test_CrystalBgGraphDisordered(unittest.TestCase):
    def setUp(self) -> None:
        ce = CheckElements.from_pymatgen_structures()
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]
        self.data0_checked = ce.check(self.data)[:10]

        gt = CrystalGraph(n_jobs=1, batch_calculate=True, batch_size=10)
        data = gt.transform(self.data0_checked)

        gen = GraphGenerator(*data, targets=None)
        self.gen = gen

    def test_sf(self):
        data0 = self.gen[0]

    def test_sf_data(self):
        gen = self.gen
        loader = MGEDataLoader(
            dataset=gen,  # torch TensorDataset format
            batch_size=3,  # 最新批数据
            shuffle=False,  # 是否随机打乱数据
            num_workers=0,  # 用于加载数据的子进程
            collate_marks=('c', 'c', 's', 'c', 'f', 'c', 'f')
        )
        for k, i in enumerate(loader):
            # if k<5:
            print("\n")
            print(k)


if __name__ == '__main__':
    unittest.main()



from pathlib import Path

import numpy as np

from featurebox.featurizers.base_transform import BaseFeature

MODULE_DIR = Path(__file__).parent.parent.parent.absolute()


# ###############bond##########################################

class BondGaussianConverter(BaseFeature):
    """
    For bond distance.
    Expand distance with Gaussian basis sit at centers and with width 0.5.
    """

    def __init__(self, centers: np.ndarray = None, width=0.5):
        """

        Args:
            centers: (np.array) centers for the Gaussian basis
            width: (float) width of Gaussian basis
        """
        if centers is None:
            centers = np.linspace(0, 5, 100)
        self.centers = centers
        self.width = width
        self.d2 = False
        super(BondGaussianConverter, self).__init__()

    def _convert(self, d: np.ndarray) -> np.ndarray:
        """
        expand distance vector d with given parameters

        Args:
            d: (1d array) distance array

        Returns:
            (matrix) N*M matrix with N the length of d and M the length of centers
        """
        d = np.array(d)

        return np.exp(-((d[:, None] - self.centers[None, :]) ** 2) / self.width ** 2)



from __future__ import division

import os.path

import numba
import numpy as np

# print('load the precomputed values from ', os.path.join(os.path.dirname(__file__), 'Wigner_coefficients.npy'))
_Wigner_coefficients = np.load(os.path.join(os.path.dirname(__file__), 'Wigner_coefficients.npy'))


@numba.njit(numba.i8(numba.i8, numba.i8, numba.i8),
            cache=True, fastmath=True, nogil=True)
def _Wigner_index(twoj, twomp, twom):
    return twoj * ((2 * twoj + 3) * twoj + 1) // 6 + (twoj + twomp) // 2 * (twoj + 1) + (twoj + twom) // 2


@numba.njit(numba.f8(numba.i8, numba.i8, numba.i8),
            cache=True, fastmath=True, nogil=True)
def _Wigner_coefficient(twoj, twomp, twom):
    return _Wigner_coefficients[_Wigner_index(twoj, twomp, twom)]


@numba.njit(numba.f8(numba.f8, numba.f8, numba.f8),
            cache=True, fastmath=True, nogil=True)
def Wigner_coefficient(j, mp, m):
    return _Wigner_coefficient(round(2 * j), round(2 * mp), round(2 * m))



import unittest

from featurebox.selection.backforward import BackForward, BackForwardStable


class MyTestCase(unittest.TestCase):
    def test_something(self):
        from sklearn.datasets import load_boston
        from sklearn.svm import SVR
        X, y = load_boston(return_X_y=True)
        svr = SVR()
        bf = BackForward(svr, primary_feature=4, random_state=1)
        new_x = bf.fit_transform(X[:,:6], y)
        bf.support_
        print(bf)
        
    def test_something2(self):
        from sklearn.datasets import load_boston
        from sklearn.svm import SVR
        X, y = load_boston(return_X_y=True)
        X=X[:,:6]
        svr = SVR()
        bf = BackForward(svr, primary_feature=4, random_state=1, refit=True)
        new_x = bf.fit_transform(X[:50], y[:50])
        test_score = bf.score(X[50:], y[50:])
        print(bf)

    def test_something3(self):

        from sklearn.datasets import load_boston
        from sklearn.svm import SVR
        from sklearn import model_selection
        X, y = load_boston(return_X_y=True)
        X=X[:,:6]
        svr = SVR()
        gd = model_selection.GridSearchCV(svr, param_grid=[{"C": [1, 10]}], n_jobs=1, cv=5)
        bf = BackForwardStable(gd, primary_feature=4, random_state=1, refit=True)
        new_x = bf.fit(X, y)
        test_score = bf.score(X, y)
        print(test_score)

    def test_something4(self):

        from sklearn.datasets import load_boston
        from sklearn.svm import SVR
        from sklearn import model_selection
        X, y = load_boston(return_X_y=True)
        X=X[:, :6]
        svr = SVR()
        bf = BackForward(svr, primary_feature=4, random_state=1, refit=True)
        new_x = bf.fit_transform(X, y)
        test_score = bf.score(X, y)


if __name__ == '__main__':
    unittest.main()
    # from sklearn.datasets import load_boston
    # from sklearn.svm import SVR
    #
    # X, y = load_boston(return_X_y=True)
    # svr = SVR()
    # bf = BackForward(svr, primary_feature=4, random_state=1, refit=True)
    # new_x = bf.fit_transform(X[:50], y[:50])
    # test_score = bf.score(X[50:], y[50:])


import unittest

import pandas as pd

from featurebox.featurizers.batch_feature import BatchFeature


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]

    def test_something2(self):
        sppm = BatchFeature(data_type="compositions", return_type="df")
        data = sppm.fit_transform([i.composition.to_reduced_dict for i in self.data0_3])

    def test_something3(self):
        sppm = BatchFeature(data_type="elements")
        aa = []
        aas = [[i.species.as_dict() for i in structure.sites] for structure in self.data0_3]
        [aa.extend(i) for i in aas]
        data = sppm.fit_transform([aa])
        data = sppm.fit_transform([aa])

    def test_something4(self):
        structures_list = self.data0_3
        sppm = BatchFeature(data_type="structures", return_type="df")
        data = sppm.fit_transform(structures_list)


if __name__ == '__main__':
    unittest.main()



# Numpy Version

from itertools import combinations, combinations_with_replacement

import numpy as np
from ase.neighborlist import NeighborList


class BehlerParrinello:
    """A class for calculating Behler-Parrinello symmetry functions.
    
    The forms of the functions are consistent with the 
    functions presented in:
    Behler, J. (2011). The Journal of Chemical Physics, 134(7), 074106.
        
    Parameters
    ----------
    symmetry_parameters: dict
        The user-defined parameters for the symmetry functions calculations.
        i.e. {'G2': {'eta': [0.1, 0.2,], 'Rs': [0.0, 0.1]}}
    Rc: float
        The symmetry functions will be calculated within this radius.
    derivative: bool
        If True, calculate the derivatives.
    stress: bool
        If True, calculate the virial stress contribution.
    """

    def __init__(self, symmetry_parameters=None, Rc=6.5,
                 derivative=True, stress=False, atom_weighted=False):

        if symmetry_parameters is None:
            symmetry_parameters = {'G2': {'eta': [0.1, 0.2, ], 'Rs': [0.0, 0.1]}}
        if atom_weighted:
            self._type = 'wACSF'
        else:
            self._type = 'BehlerParrinello'

        # Set up the symmetry parameters keywords. If a string are not in the
        # keyword, code will return an error.
        G2_keywords = ['Rs', 'eta']
        G4_keywords = ['Rs', 'eta', 'lambda', 'zeta']
        G5_keywords = G4_keywords

        # Setting up parameters for each of the symmetry function type.
        self.G2_parameters = None
        self.G4_parameters = None
        self.G5_parameters = None

        for key, value in symmetry_parameters.items():
            if key == 'G2':
                self.G2_parameters = {'eta': [1], 'Rs': np.array([0.])}
                for key0, value0 in value.items():
                    if key0 in G2_keywords:
                        self.G2_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G2_keywords}"
                        raise NotImplementedError(msg)

            elif key == 'G4':
                self.G4_parameters = {'Rs': np.array([0.]),
                                      'eta': [1], 'zeta': [1], 'lambda': [1]}
                for key0, value0 in value.items():
                    if key0 in G4_keywords:
                        self.G4_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G4_keywords}"
                        raise NotImplementedError(msg)

            elif key == 'G5':
                self.G5_parameters = {'Rs': np.array([0.]),
                                      'eta': [1], 'zeta': [1], 'lambda': [1]}
                for key0, value0 in value.items():
                    if key0 in G5_keywords:
                        self.G5_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G5_keywords}"
                        raise NotImplementedError(msg)

        self.Rc = Rc
        self.derivative = derivative
        self.stress = stress

    def calculate(self, crystal, system=None):
        """The symmetry functions are obtained through this `calculate` method.
        
        Parameters
        ----------
        crystal: object
            ASE Structure object.
        system: list
            A list of the crystal structures system. 
            All elements in the list have to be integer. For example, the
            system of crystal structure is NaCl system. Then, system should be
            pass as [11, 17]
            
        Returns
        -------
        all_G: dict
            The user-defined symmetry functions that represent the crystal.
            Currently, there are 3 types of symmetry functions implemented. 
            Here are the order of the descriptors are printed out based on
            their symmetry parameters:
            - G2: ["element", "Rs", "eta"]
            - G4: ["pair_elements", "eta", "lambda", "zeta"]
            - G5: ["pair_elements", "eta", "lambda", "zeta"]
        """
        self.crystal = crystal
        atomic_numbers = np.array(crystal.get_atomic_numbers())
        vol = crystal.get_volume()

        if system is None:
            type_set1 = create_type_set(atomic_numbers, 1)  # l
            type_set2 = create_type_set(atomic_numbers, 2)  # l
        else:
            system = np.array(system, dtype=int)
            type_set1 = create_type_set(system, 1)  # l
            type_set2 = create_type_set(system, 2)  # l

        # Initialize dict for the symmetry functions
        self.all_G = {'x': [],
                      'elements': []}
        if self.derivative:
            self.all_G['dxdr'] = None
            self.all_G['seq'] = None
        if self.stress:
            self.all_G['rdxdr'] = None

        # Obtain neighbors info.
        rc = [self.Rc / 2] * len(self.crystal)
        neighbors = NeighborList(rc, self_interaction=False, bothways=True,
                                 skin=0.0)
        neighbors.update(crystal)

        for i in range(len(crystal)):
            element = crystal.get_chemical_symbols()[i]
            indices, offsets = neighbors.get_neighbors(i)

            assert len(indices) > 0, \
                f"There's no neighbor for this structure at Rc = {self.Rc} A."

            Ri = crystal.get_positions()[i]
            total_neighbors = len(indices)  # total number of neighbors of atom i

            Rj = np.zeros([total_neighbors, 3])
            Dij = np.zeros(total_neighbors)
            IDs = np.zeros(total_neighbors, dtype=int)
            jks = np.array(list(combinations(range(total_neighbors), 2)))

            count = 0
            for j, offset in zip(indices, offsets):
                Rj[count, :] = crystal.positions[j] + np.dot(offset, crystal.get_cell())
                Dij[count] = np.sqrt(sum((Rj[count, :] - Ri) ** 2))
                IDs[count] = j
                count += 1

            Rij = Rj - Ri

            Gi = []
            GiPrime = None
            GiPrime_seq = None

            if self.G2_parameters is not None:
                G2i = calculate_G2(Dij, IDs, atomic_numbers, type_set1,
                                   self.Rc, self.G2_parameters, self._type)
                Gi = np.append(Gi, G2i)

                if self.derivative:
                    seq, G2iP, rG2iP = calculate_G2Prime(Rij, Ri, i, IDs, atomic_numbers,
                                                         type_set1, self.Rc,
                                                         self.G2_parameters, self._type)
                    if GiPrime is None:
                        GiPrime = G2iP  # [N1, l, 3]
                        GiPrime_seq = seq
                        if self.stress:
                            rGiPrime = rG2iP
                    else:
                        GiPrime_seq = np.append(GiPrime_seq, seq, axis=0)
                        GiPrime = np.append(GiPrime, G2iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG2iP, axis=1)

            if self.G4_parameters is not None:
                G4i = calculate_G4(Rij, IDs, jks, atomic_numbers, type_set2,
                                   self.Rc, self.G4_parameters, self._type)
                Gi = np.append(Gi, G4i)

                if self.derivative:
                    seq, G4iP, rG4iP = calculate_G4Prime(Rij, Ri, i, IDs, jks, atomic_numbers,
                                                         type_set2, self.Rc,
                                                         self.G4_parameters, self._type)
                    if GiPrime is None:
                        GiPrime = G4iP
                        GiPrime_seq = seq
                        if self.stress:
                            rGiPrime = rG4iP
                    else:
                        # GiPrime_seq = np.append(GiPrime_seq, seq, axis=0)
                        GiPrime = np.append(GiPrime, G4iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG4iP, axis=1)
                # print(seq.shape, G4iP.shape, rG4iP.shape)
                # print(GiPrime_seq.shape, GiPrime.shape, rGiPrime.shape)

            if self.G5_parameters is not None:
                G5i = calculate_G5(Rij, IDs, jks, atomic_numbers, type_set2,
                                   self.Rc, self.G5_parameters, self._type)
                Gi = np.append(Gi, G5i)

                if self.derivative:
                    seq, G5iP, rG5iP = calculate_G5Prime(Rij, Ri, i, IDs, jks, atomic_numbers,
                                                         type_set2, self.Rc,
                                                         self.G5_parameters, self._type)
                    if GiPrime is None:
                        GiPrime = G5iP
                        GiPrime_seq = seq
                        if self.stress:
                            rGiPrime = rG5iP
                    else:
                        GiPrime = np.append(GiPrime, G5iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG5iP, axis=1)

            self.all_G['x'].append(Gi)
            self.all_G['elements'].append(element)
            if self.derivative:
                if self.all_G['seq'] is None:
                    self.all_G['seq'] = GiPrime_seq
                    self.all_G['dxdr'] = GiPrime
                    if self.stress:
                        self.all_G['rdxdr'] = rGiPrime
                else:
                    self.all_G['seq'] = np.append(self.all_G['seq'], GiPrime_seq, axis=0)
                    self.all_G['dxdr'] = np.append(self.all_G['dxdr'], GiPrime, axis=0)
                    if self.stress:
                        self.all_G['rdxdr'] = np.append(self.all_G['rdxdr'], rGiPrime, axis=0)

        self.all_G['x'] = np.asarray(self.all_G['x'])
        if self.derivative:
            self.all_G['dxdr'] = np.asarray(self.all_G['dxdr'])
            self.all_G['seq'] = np.asarray(self.all_G['seq'])
            if self.stress:
                self.all_G['rdxdr'] = -np.asarray(self.all_G['rdxdr']) / vol
            else:
                self.all_G['rdxdr'] = None
        return self.all_G


########################### Symmetry Functions ################################


def calculate_G2(Dij, IDs, atomic_numbers, type_set, Rc, parameters, Gtype):
    """Calculate G2 symmetry function for a center atom i.
    
    G2 function describes the radial feature of atoms in a crystal structure
    given a cutoff radius. 

    Parameters
    ----------
    Dij: float array [m]
        The array of distances for a given atom.
    IDs: int array [m]
        The indices of neighbors centering about atom i.
    atomic_numbers: int array [N]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (n1)
            The shift from the center of the G2 symmetry function.
        etas: float array (n2)
             eta parameters of G2 symmetry function.
    
    Returns
    -------
    G2: 1D float array [d = n1*n2*l]
        G2 symmetry value.
    """
    Rs = parameters['Rs']  # n1
    etas = parameters['eta']  # n2
    n1, n2, m, l = len(Rs), len(etas), len(Dij), len(type_set)

    d20 = (Dij - Rs[:, np.newaxis]) ** 2  # n1*m
    term = np.exp(np.einsum('i,jk->ijk', -etas, d20))  # n2*n1*m
    results = np.einsum('ijk,k->ijk', term, Cosine(Dij, Rc))  # n2*n1*m
    results = results.reshape([n1 * n2, m])

    # Decompose G2 by species
    if Gtype == 'wACSF':
        G2 = np.zeros([n1 * n2])
        j_ids = atomic_numbers[IDs]
        for id, j_type in enumerate(type_set):
            ids = select_rows(j_ids, j_type)
            Z = j_ids[ids][0]
            G2 += Z * np.einsum('ij->i', results[:, ids])

    else:
        G2 = np.zeros([n1 * n2 * l])
        j_ids = atomic_numbers[IDs]
        for id, j_type in enumerate(type_set):
            ids = select_rows(j_ids, j_type)
            G2[id::l] = np.sum(results[:, ids], axis=1)

    return G2


def calculate_G2Prime(Rij, Ri, i, IDs, atomic_numbers, type_set, Rc, parameters, Gtype):
    """Calculate the derivative of G2 symmetry function for atom i.
    
    Parameters
    ----------
    Rij: float array [m, 3]
        The vector distances of atom i to neighbors js.
    i: int
        The i-th atom center.
    IDs: int array [m]
        The indices of neighbors centering about atom i.
    atomic_numbers: int array [N]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (n1)
            The shift from the center of the G2 symmetry function.
        etas: float array (n2)
             eta parameters of G2 symmetry function.
    
    Returns
    -------
    G2Prime: 1D float array [m, (n1*n2*l), 3]
        The derivative of G2 symmetry value at i-th atom. m is the index of the
        atom that force is acting on.
        
    """
    etas = parameters['eta']
    Rs = parameters['Rs']
    n1, n2, m, l = len(Rs), len(etas), len(Rij), len(type_set)
    js, N = IDs, len(atomic_numbers)
    # QZ: the block to get the unique neighbors
    unique_js = np.unique(IDs)
    if i not in unique_js:
        unique_js = np.append(i, unique_js)
    seq = i * np.ones([len(unique_js), 2], dtype=int)
    seq[:, 1] = unique_js
    N1 = len(unique_js)
    _i = np.where(unique_js == i)[0][0]

    ij_list = i * np.ones([len(js), 2], dtype=int)
    ij_list[:, 1] = js

    R2ij = np.sum(Rij ** 2., axis=1)
    R1ij = np.sqrt(R2ij)
    dij = R1ij - Rs[:, np.newaxis]
    dij2 = (dij) ** 2
    term1 = np.exp(np.einsum('i,jk->ijk', -etas, dij2))  # n2*n1*m
    term21 = CosinePrime(R1ij, Rc)  # m
    _term22 = np.einsum('ij,j->ij', dij, Cosine(R1ij, Rc))  # n1*m
    term22 = 2 * np.einsum('i,jk->ijk', etas, _term22)  # n2*n1*m
    term2 = term21 - term22
    term_1_and_2 = term1 * term2
    term_1_and_2 = term_1_and_2.reshape([n1 * n2, m])

    dRij_dRm = np.zeros([m, 3, N1])
    for mm, _m in enumerate(unique_js):
        mm_list = _m * np.ones([m, 1], dtype=int)
        dRij_dRm[:, :, mm] = dRij_dRm_norm(Rij, np.hstack((ij_list, mm_list)))

    # [d, m, 3, N1]
    G2ip0 = np.einsum('ij,jkl->ijkl', term_1_and_2, dRij_dRm)

    rG2ip0 = np.zeros([n1 * n2, len(ij_list), 3, N1, 3])
    for mm, ij in enumerate(ij_list):
        # j = ij[1]
        _j = np.where(unique_js == ij[1])[0][0]
        tmp = G2ip0[:, mm, :, :]  # S,N1,3 -> S,3 * 3 -> S*3*3
        rG2ip0[:, mm, :, _i, :] += np.einsum('ij,k->ijk', tmp[:, :, _i], Ri)
        rG2ip0[:, mm, :, _j, :] += np.einsum('ij,k->ijk', tmp[:, :, _j], Rij[mm] + Ri)

    # Decompose G2 Prime by species
    if Gtype == 'wACSF':
        j_ids = atomic_numbers[IDs]

        G2Prime = np.zeros([N1, n1 * n2, 3])  # This is the final output
        rG2Prime = np.zeros([N1, n1 * n2, 3, 3])  # This is the final output

        for id, j_type in enumerate(type_set):
            ids = select_rows(j_ids, j_type)
            Z = j_ids[ids][0]
            G2Prime += Z * np.einsum('ijkl->lik', G2ip0[:, ids, :, :])
            rG2Prime += Z * np.einsum('ijklm->likm', rG2ip0[:, ids, :, :, :])

    else:
        j_ids = atomic_numbers[js]
        G2Prime = np.zeros([N1, n1 * n2 * l, 3])  # This is the final output
        rG2Prime = np.zeros([N1, n1 * n2 * l, 3, 3])  # This is the final output

        for id, j_type in enumerate(type_set):
            ids = select_rows(j_ids, j_type)
            G2Prime[:, id::l, :] += np.einsum('ijkl->lik', G2ip0[:, ids, :, :])
            rG2Prime[:, id::l, :, :] += np.einsum('ijklm->likm', rG2ip0[:, ids, :, :, :])

    return seq, G2Prime, rG2Prime


def calculate_G4(Rij, IDs, jks, atomic_numbers, type_set, Rc, parameters, Gtype):
    """Calculate G4 symmetry function for a given atom i.
    
    G4 function also describes the angular feature of atoms in a crystal 
    structure given a cutoff radius.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    IDs: int array [m]
        The indices of neighbors centering about atom i.
    jks: int array [m*(m-1)/2, 2]
        The list of [j,k] pairs
    atomic_numbers: int array [N]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (n1)
        etas: float array (n2)
        lamBdas: float array (n3)
        zetas: float array (n4)
        
    Returns
    -------
    G4: array [d = n1*n2*n3*n4*l]
        G4 symmetry value with d members.
    """
    Rs = parameters['Rs']
    etas = parameters['eta']
    zetas = parameters['zeta']
    lamBdas = parameters['lambda']
    n1, n2, n3, n4, l = len(Rs), len(etas), len(lamBdas), len(zetas), len(type_set)
    jk = len(jks)  # m1

    rij = Rij[jks[:, 0]]  # [m1, 3]
    rik = Rij[jks[:, 1]]  # [m1, 3]
    rjk = rik - rij  # [m1, 3]
    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    R2jk0 = np.sum(rjk ** 2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # m1
    R1ik0 = np.sqrt(R2ik0)  # m1
    R1jk0 = np.sqrt(R2jk0)  # m1
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2jk = R2jk0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R1ij = R1ij0 - Rs[:, np.newaxis]  # n1*m1
    R1ik = R1ik0 - Rs[:, np.newaxis]  # n1*m1
    R1jk = R1jk0 - Rs[:, np.newaxis]  # n1*m1

    powers = 2. ** (1. - zetas)  # n4
    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # m1 array
    term1 = 1. + np.einsum('i,j->ij', lamBdas, cos_ijk)  # n3*m1

    zetas1 = zetas.repeat(n3 * jk).reshape([n4, n3, jk])  # n4*n3*m1
    term2 = np.power(term1, zetas1)  # n4*n3*m1
    term3 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2jk + R2ik)))  # n2*n1*m1
    term4 = Cosine(R1ij0, Rc) * Cosine(R1ik0, Rc) * Cosine(R1jk0, Rc)  # m1
    term5 = np.einsum('ijk,lmk->ijlmk', term2, term3)  # n4*n3*n2*n1*m1
    term6 = np.einsum('ijkml,l->ijkml', term5, term4)  # n4*n3*n2*n1*m1
    results = np.einsum('i,ijkml->ijkml', powers, term6)  # n4*n3*n2*n1*m1
    results = results.reshape([n1 * n2 * n3 * n4, jk])

    # Decompose G4 by species
    if Gtype == 'wACSF':
        G4 = np.zeros([n1 * n2 * n3 * n4])
        jk_ids = atomic_numbers[IDs[jks]]

        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            Z1Z2 = np.prod(jk_ids[ids[0]])
            G4 += Z1Z2 * np.sum(results[:, ids], axis=1)

    else:
        G4 = np.zeros([n1 * n2 * n3 * n4 * l])
        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            G4[id::l] = np.sum(results[:, ids], axis=1)

    return G4


def calculate_G5(Rij, IDs, jks, atomic_numbers, type_set, Rc, parameters, Gtype):
    """Calculate G5 symmetry function for a given atom i.
    
    G5 function also describes the angular feature of atoms in a crystal 
    structure given a cutoff radius.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    IDs: int array [m]
        The indices of neighbors centering about atom i.
    jks: int array [m*(m-1)/2, 2]
        The list of [j,k] pairs
    atomic_numbers: int array [N]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (n1)
        etas: float array (n2)
        lamBdas: float array (n3)
        zetas: float array (n4)
        
    Returns
    -------
    G5: array [d = n1*n2*n3*n4*l]
        G5 symmetry value with d members.
    """
    Rs = parameters['Rs']
    etas = parameters['eta']
    zetas = parameters['zeta']
    lamBdas = parameters['lambda']
    n1, n2, n3, n4, l = len(Rs), len(etas), len(lamBdas), len(zetas), len(type_set)
    jk = len(jks)  # m1

    rij = Rij[jks[:, 0]]  # [m1, 3]
    rik = Rij[jks[:, 1]]  # [m1, 3]
    # rjk = rik - rij     # [m1, 3]
    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)

    # R2jk0 = np.sum(rjk**2., axis=1)

    R1ij0 = np.sqrt(R2ij0)  # m1
    R1ik0 = np.sqrt(R2ik0)  # m1
    # R1jk0 = np.sqrt(R2jk0) # m1

    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # n1*m1

    # R2jk = R2jk0 - Rs[:, np.newaxis]**2  # n1*m1

    R1ij = R1ij0 - Rs[:, np.newaxis]  # n1*m1
    R1ik = R1ik0 - Rs[:, np.newaxis]  # n1*m1

    # R1jk = R1jk0 - Rs[:, np.newaxis] # n1*m1

    powers = 2. ** (1. - zetas)  # n4
    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # m1 array
    term1 = 1. + np.einsum('i,j->ij', lamBdas, cos_ijk)  # n3*m1

    zetas1 = zetas.repeat(n3 * jk).reshape([n4, n3, jk])  # n4*n3*m1
    term2 = np.power(term1, zetas1)  # n4*n3*m1
    term3 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2ik)))  # n2*n1*m1
    term4 = Cosine(R1ij0, Rc) * Cosine(R1ik0, Rc)  # * Cosine(R1jk0, Rc) # m1
    term5 = np.einsum('ijk,lmk->ijlmk', term2, term3)  # n4*n3*n2*n1*m1
    term6 = np.einsum('ijkml,l->ijkml', term5, term4)  # n4*n3*n2*n1*m1
    results = np.einsum('i,ijkml->ijkml', powers, term6)  # n4*n3*n2*n1*m1
    results = results.reshape([n1 * n2 * n3 * n4, jk])

    # Decompose G5 by species
    if Gtype == 'wACSF':
        G5 = np.zeros([n1 * n2 * n3 * n4])
        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            Z1Z2 = np.prod(jk_ids[ids[0]])
            G5 += Z1Z2 * np.sum(results[:, ids], axis=1)

    else:
        G5 = np.zeros([n1 * n2 * n3 * n4 * l])
        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            G5[id::l] = np.sum(results[:, ids], axis=1)

    return G5


def calculate_G4Prime(Rij, Ri, i, IDs, jks, atomic_numbers, type_set, Rc,
                      parameters, Gtype):
    """Calculate G4 symmetry function for a given atom i.
    
    G4 function also describes the angular feature of atoms in a crystal 
    structure given a cutoff radius.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    i: int
        The i-th atom center.    
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    jks: int array [j*k, 2]
        The list of [j,k] pairs
    atomic_numbers: int array [j]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        zetas: float array (n1)
        lamBdas: float array (n2)
        etas: float array (n3)
        
    Returns

        The derivative of G4 symmetry value at i-th atom. m is the index of the
        atom that force is acting on.
    """
    Rs = parameters['Rs']
    etas = parameters['eta']
    zetas = parameters['zeta']
    lamBdas = parameters['lambda']
    n1, n2, n3, n4, l = len(Rs), len(etas), len(lamBdas), len(zetas), len(type_set)
    N, jk = len(atomic_numbers), len(jks)

    # QZ: the block to get the unique neighbors
    unique_js = np.unique(IDs)
    if i not in unique_js:
        unique_js = np.append(i, unique_js)
    seq = i * np.ones([len(unique_js), 2], dtype=int)
    seq[:, 1] = unique_js
    N1 = len(unique_js)
    _i = np.where(unique_js == i)[0][0]

    ijk_list = i * np.ones([jk, 3], dtype=int)
    ijk_list[:, 1] = IDs[jks[:, 0]]
    ijk_list[:, 2] = IDs[jks[:, 1]]

    rij = Rij[jks[:, 0]]  # [m1, 3]
    rik = Rij[jks[:, 1]]  # [m1, 3]
    rjk = rik - rij  # [m1, 3]

    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    R2jk0 = np.sum(rjk ** 2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # m1
    R1ik0 = np.sqrt(R2ik0)  # m1
    R1jk0 = np.sqrt(R2jk0)  # m1
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2jk = R2jk0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R1ij = R1ij0 - Rs[:, np.newaxis]  # n1*m1
    R1ik = R1ik0 - Rs[:, np.newaxis]  # n1*m1
    R1jk = R1jk0 - Rs[:, np.newaxis]  # n1*m1

    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # m1 array

    dfcij = CosinePrime(R1ij0, Rc)
    dfcjk = CosinePrime(R1jk0, Rc)
    dfcik = CosinePrime(R1ik0, Rc)
    fcij = Cosine(R1ij0, Rc)
    fcjk = Cosine(R1jk0, Rc)
    fcik = Cosine(R1ik0, Rc)

    powers = 2. ** (1. - zetas)  # n4
    term1 = 1. + np.einsum('i,j->ij', lamBdas, cos_ijk)  # n3*m1
    zetas1 = zetas.repeat(n3 * jk).reshape([n4, n3, jk])  # n4*n3*m1
    term2 = np.power(term1, zetas1 - 1)  # n4*n3*m1
    g41 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2jk + R2ik)))  # n2*n1*m1
    g41 = np.einsum('ijk,lmk->ijlmk', term2, g41)  # n4*n3*n2**n1*m1
    g41 = np.einsum('i,ijklm->ijklm', powers, g41)  # n4*n3*n2*n1*m1

    lamBda_zeta = np.einsum('i,j->ij', zetas, lamBdas)  # n4*n3
    (dRij_dRm, dRik_dRm, dRjk_dRm) = dRijk_dRm(rij, rik, rjk, ijk_list, unique_js)  # m1*3*N1
    Rijk_dRm = np.einsum('i,ijk->ijk', R1ij0, dRij_dRm) + \
               np.einsum('i,ijk->ijk', R1ik0, dRik_dRm) + \
               np.einsum('i,ijk->ijk', R1jk0, dRjk_dRm)
    dcos = dcosijk_dRm(rij, rik, ijk_list, unique_js, dRij_dRm, dRik_dRm)
    dcos = np.einsum('ij,klm->ijklm', lamBda_zeta, dcos)  # n4*n3*3*m1*N1
    dcos = np.broadcast_to(dcos, (n2,) + (n4, n3, jk, 3, N1))
    dcos = np.transpose(dcos, (1, 2, 0, 3, 4, 5))
    cost = np.einsum('i,jk->jik', 2 * etas, term1)  # n3*n2*m1
    cost = np.einsum('ijk,klm->ijklm', cost, Rijk_dRm)  # n3*n2*m1*3*N1
    cost = np.broadcast_to(cost, (n4,) + (n3, n2, jk, 3, N1))
    g42 = np.einsum('l,ijklmn->ijklmn', fcij * fcik * fcjk, dcos - cost)  # n4*n3*n2*m*3*N1

    g43 = np.einsum('i,ijk->ijk', dfcij * fcik * fcjk, dRij_dRm) + \
          np.einsum('i,ijk->ijk', fcij * dfcik * fcjk, dRik_dRm) + \
          np.einsum('i,ijk->ijk', fcij * fcik * dfcjk, dRjk_dRm)
    g43 = np.einsum('ij,jkl->ijkl', term1, g43)  # n3*m1*3*N1
    g43 = np.broadcast_to(g43, (n4, n2,) + (n3, jk, 3, N1))
    g43 = np.transpose(g43, (0, 2, 1, 3, 4, 5))

    # [n4, n3, n2, n1, m] * [n4, n3, n2, m, 3, N1] -> [n4, n3, n2, n1, m, 3, N1] -> [S, m, 3, N1] 
    G4ip0 = np.einsum('ijklm, ijkmno->ijklmno', g41, g42 + g43, \
                      optimize='greedy').reshape([n1 * n2 * n3 * n4, jk, 3, N1])
    # [S, m, 3, N1] * [m, 3] -> [S, m, 3, N1, 3] 
    # partition the dxdr to each i, j, k
    rG4ip0 = np.zeros([n1 * n2 * n3 * n4, len(ijk_list), 3, N1, 3])
    for mm, ijk in enumerate(ijk_list):
        # j,k = ijk[1], ijk[2]
        _j = np.where(unique_js == ijk[1])[0][0]
        _k = np.where(unique_js == ijk[2])[0][0]
        tmp = G4ip0[:, mm, :, :]  # S,N,3 -> S,3 * 3 -> S*3*3
        rG4ip0[:, mm, :, _i, :] += np.einsum('ij,k->ijk', tmp[:, :, _i], Ri)
        rG4ip0[:, mm, :, _j, :] += np.einsum('ij,k->ijk', tmp[:, :, _j], rij[mm] + Ri)
        rG4ip0[:, mm, :, _k, :] += np.einsum('ij,k->ijk', tmp[:, :, _k], rik[mm] + Ri)

    # Decompose G4 Prime by species
    if Gtype == 'wACSF':
        G4Prime = np.zeros([N1, n1 * n2 * n3 * n4, 3])
        rG4Prime = np.zeros([N1, n1 * n2 * n3 * n4, 3, 3])

        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            Z1Z2 = np.prod(jk_ids[ids[0]])
            G4Prime += Z1Z2 * np.einsum('ijkl->lik', G4ip0[:, ids, :, :])
            rG4Prime += Z1Z2 * np.einsum('ijklm->likm', rG4ip0[:, ids, :, :, :])

    else:
        G4Prime = np.zeros([N1, n1 * n2 * n3 * n4 * l, 3])
        rG4Prime = np.zeros([N1, n1 * n2 * n3 * n4 * l, 3, 3])

        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            G4Prime[:, id::l, :] += np.einsum('ijkl->lik', G4ip0[:, ids, :, :])
            rG4Prime[:, id::l, :, :] += np.einsum('ijklm->likm', rG4ip0[:, ids, :, :, :])

    return seq, G4Prime, rG4Prime


def calculate_G5Prime(Rij, Ri, i, IDs, jks, atomic_numbers, type_set, Rc,
                      parameters, Gtype):
    """Calculate G5 symmetry function for a given atom i.
    
    G5 function also describes the angular feature of atoms in a crystal 
    structure given a cutoff radius.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    i: int
        The i-th atom center.    
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    jks: int array [j*k, 2]
        The list of [j,k] pairs
    atomic_numbers: int array [j]
        The elemental indices for atoms within the unitcell.
        e.g. [11, 11, 11, 11, 17, 17, 17, 17]
    Rc: float
        The cutoff radius.
    parameters: dict
        zetas: float array (n1)
        lamBdas: float array (n2)
        etas: float array (n3)
        
    Returns
    -------
    G5Prime: array [m, d, 3]
        The derivative of G5 symmetry value at i-th atom. m is the index of the
        atom that force is acting on.
    """
    Rs = parameters['Rs']
    etas = parameters['eta']
    zetas = parameters['zeta']
    lamBdas = parameters['lambda']
    n1, n2, n3, n4, l = len(Rs), len(etas), len(lamBdas), len(zetas), len(type_set)
    N, jk = len(atomic_numbers), len(jks)

    # QZ: the block to get the unique neighbors
    unique_js = np.unique(IDs)
    if i not in unique_js:
        unique_js = np.append(i, unique_js)
    seq = i * np.ones([len(unique_js), 2], dtype=int)
    seq[:, 1] = unique_js
    N1 = len(unique_js)
    _i = np.where(unique_js == i)[0][0]

    ijk_list = i * np.ones([jk, 3], dtype=int)
    ijk_list[:, 1] = IDs[jks[:, 0]]
    ijk_list[:, 2] = IDs[jks[:, 1]]

    rij = Rij[jks[:, 0]]  # [m1, 3]
    rik = Rij[jks[:, 1]]  # [m1, 3]
    rjk = rik - rij  # [m1, 3]

    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    # R2jk0 = np.sum(rjk**2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # m1
    R1ik0 = np.sqrt(R2ik0)  # m1
    # R1jk0 = np.sqrt(R2jk0) # m1
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # n1*m1
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # n1*m1
    # R2jk = R2jk0 - Rs[:, np.newaxis]**2  # n1*m1
    R1ij = R1ij0 - Rs[:, np.newaxis]  # n1*m1
    R1ik = R1ik0 - Rs[:, np.newaxis]  # n1*m1
    # R1jk = R1jk0 - Rs[:, np.newaxis] # n1*m1

    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # m1 array

    dfcij = CosinePrime(R1ij0, Rc)
    # dfcjk = CosinePrime(R1jk0, Rc)
    dfcik = CosinePrime(R1ik0, Rc)
    fcij = Cosine(R1ij0, Rc)
    # fcjk = Cosine(R1jk0, Rc)
    fcik = Cosine(R1ik0, Rc)

    powers = 2. ** (1. - zetas)  # n4
    term1 = 1. + np.einsum('i,j->ij', lamBdas, cos_ijk)  # n3*m1
    zetas1 = zetas.repeat(n3 * jk).reshape([n4, n3, jk])  # n4*n3*m1
    term2 = np.power(term1, zetas1 - 1)  # n4*n3*m1
    g51 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2ik)))  # n2*n1*m1
    g51 = np.einsum('ijk,lmk->ijlmk', term2, g51)  # n4*n3*n2**n1*m1
    g51 = np.einsum('i,ijklm->ijklm', powers, g51)  # n4*n3*n2*n1*m1

    lamBda_zeta = np.einsum('i,j->ij', zetas, lamBdas)  # n4*n3
    (dRij_dRm, dRik_dRm, dRjk_dRm) = dRijk_dRm(rij, rik, rjk, ijk_list, unique_js)  # m1*3*N1
    Rijk_dRm = np.einsum('i,ijk->ijk', R1ij0, dRij_dRm) + \
               np.einsum('i,ijk->ijk', R1ik0, dRik_dRm)
    # np.einsum('i,ijk->ijk', R1jk0, dRjk_dRm)
    dcos = dcosijk_dRm(rij, rik, ijk_list, unique_js, dRij_dRm, dRik_dRm)
    dcos = np.einsum('ij,klm->ijklm', lamBda_zeta, dcos)  # n4*n3*3*m1*N1
    dcos = np.broadcast_to(dcos, (n2,) + (n4, n3, jk, 3, N1))
    dcos = np.transpose(dcos, (1, 2, 0, 3, 4, 5))
    cost = np.einsum('i,jk->jik', 2 * etas, term1)  # n3*n2*m1
    cost = np.einsum('ijk,klm->ijklm', cost, Rijk_dRm)  # n3*n2*m1*3*N1
    cost = np.broadcast_to(cost, (n4,) + (n3, n2, jk, 3, N1))
    g52 = np.einsum('l,ijklmn->ijklmn', fcij * fcik, dcos - cost)  # n4*n3*n2*m*3*N1

    g53 = np.einsum('i,ijk->ijk', dfcij * fcik, dRij_dRm) + \
          np.einsum('i,ijk->ijk', fcij * dfcik, dRik_dRm)
    g53 = np.einsum('ij,jkl->ijkl', term1, g53)  # n3*m1*3*N1
    g53 = np.broadcast_to(g53, (n4, n2,) + (n3, jk, 3, N1))
    g53 = np.transpose(g53, (0, 2, 1, 3, 4, 5))

    G5ip0 = np.einsum('ijklm, ijkmno->ijklmno', g51, g52 + g53, \
                      optimize='greedy').reshape([n1 * n2 * n3 * n4, jk, 3, N1])

    # [S, m, 3, N1] * [m, 3] -> [S, m, 3, N1, 3] 
    # partition the dxdr to each i, j, k
    rG5ip0 = np.zeros([n1 * n2 * n3 * n4, len(ijk_list), 3, N1, 3])
    for mm, ijk in enumerate(ijk_list):
        # j,k = ijk[1], ijk[2]
        _j = np.where(unique_js == ijk[1])[0][0]
        _k = np.where(unique_js == ijk[2])[0][0]
        tmp = G5ip0[:, mm, :, :]  # S,N,3 -> S,3 * 3 -> S*3*3
        rG5ip0[:, mm, :, _i, :] += np.einsum('ij,k->ijk', tmp[:, :, _i], Ri)
        rG5ip0[:, mm, :, _j, :] += np.einsum('ij,k->ijk', tmp[:, :, _j], rij[mm] + Ri)
        rG5ip0[:, mm, :, _k, :] += np.einsum('ij,k->ijk', tmp[:, :, _k], rik[mm] + Ri)

    # Decompose G4 Prime by species
    if Gtype == 'wACSF':
        G5Prime = np.zeros([N1, n1 * n2 * n3 * n4, 3])
        rG5Prime = np.zeros([N1, n1 * n2 * n3 * n4, 3, 3])

        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            Z1Z2 = np.prod(jk_ids[ids[0]])
            G5Prime += Z1Z2 * np.einsum('ijkl->lik', G5ip0[:, ids, :, :])
            rG5Prime += Z1Z2 * np.einsum('ijklm->likm', rG5ip0[:, ids, :, :, :])

    else:
        G5Prime = np.zeros([N1, n1 * n2 * n3 * n4 * l, 3])
        rG5Prime = np.zeros([N1, n1 * n2 * n3 * n4 * l, 3, 3])

        jk_ids = atomic_numbers[IDs[jks]]
        for id, jk_type in enumerate(type_set):
            ids = select_rows(jk_ids, jk_type)
            G5Prime[:, id::l, :] += np.einsum('ijkl->lik', G5ip0[:, ids, :, :])
            rG5Prime[:, id::l, :, :] += np.einsum('ijklm->likm', rG5ip0[:, ids, :, :, :])

    return seq, G5Prime, rG5Prime


def dRij_dRm_norm(Rij, ijm_list):
    """Calculate the derivative of Rij norm w. r. t. atom m. This term affects 
    only on i and j.
    
    Parameters
    ----------
    Rij : array [j, 3] or [j*k, 3]
        The vector distances of atom i to atom j.
    ijm_list: array [j, 3] or [j*k, 3]
        Id list of center atom i, neighbors atom j, and atom m.
    
    Returns
    -------
    dRij_m: array [j, 3] or [j*k, 3]
        The derivative of pair atoms w.r.t. atom m in x, y, z directions.
    """
    dRij_m = np.zeros([len(Rij), 3])
    R1ij = np.linalg.norm(Rij, axis=1).reshape([len(Rij), 1])
    l1 = (ijm_list[:, 2] == ijm_list[:, 0])
    dRij_m[l1, :] = -Rij[l1] / R1ij[l1]
    l2 = (ijm_list[:, 2] == ijm_list[:, 1])
    dRij_m[l2, :] = Rij[l2] / R1ij[l2]
    l3 = (ijm_list[:, 0] == ijm_list[:, 1])
    dRij_m[l3, :] = 0

    return dRij_m


def dcosijk_dRm(Rij, Rik, ijk_list, m_list, dRij_dRm, dRik_dRm):
    """Calculate the derivative of cosine_ijk function w. r. t. atom m.
    m must belong to one of the ijks. Otherwise, the derivative is zero.
    If the input Rij and Rik are (j*k)*3 dimensions, the output will be 
    (j*k)*3*3. The extra dimension comes from looping over {i, j, k}.
    
    Parameters
    ----------
    Rij: array [j*k, 3]
        The vector distances of atom i to neighbors js.
    Rik: array [j*k, 3]
        The vector distances of atom i to neighbors ks, where j != k.
        
    Returns
    -------
    Derivative of cosine dot product w.r.t. the radius of an atom m,
    The atom m has to be in the an array with 3 indices: i, j, and k.
    """
    m = dRik_dRm.shape[-1]
    Dij = np.linalg.norm(Rij, axis=1)
    Dik = np.linalg.norm(Rik, axis=1)
    rDijDik = 1 / Dij / Dik  # jk array
    Rij_Rik = np.sum(Rij * Rik, axis=1)  # jk array
    ij_list = ijk_list[:, [0, 1]]
    ik_list = ijk_list[:, [0, 2]]
    dcos = np.zeros([len(Rij), 3, m])

    # for mm in range(m):
    for mm, _m in enumerate(m_list):
        mm_list = _m * np.ones([len(Rij), 1], dtype=int)
        # mm_list = mm * np.ones([len(Rij), 1], dtype=int)
        dRij_dRm_v = dRij_dRm_vector(np.append(ij_list, mm_list, axis=1))
        dRik_dRm_v = dRij_dRm_vector(np.append(ik_list, mm_list, axis=1))
        term10 = np.einsum('i,ij->ij', rDijDik, Rik)  # jk*3 arrray
        dcos[:, :, mm] += np.einsum('ij,i->ij', term10, dRij_dRm_v)
        term20 = np.einsum('i,ij->ij', rDijDik, Rij)  # jk*3 array
        dcos[:, :, mm] += np.einsum('ij,i->ij', term20, dRik_dRm_v)
        term30 = Rij_Rik * rDijDik / Dij  # jk*3
        dcos[:, :, mm] -= np.einsum('i,ij->ij', term30, dRij_dRm[:, :, mm])
        term40 = Rij_Rik * rDijDik / Dik  # jk*3
        dcos[:, :, mm] -= np.einsum('i,ij->ij', term40, dRik_dRm[:, :, mm])  # jk*3*m

    return dcos


def dRij_dRm_vector(ijm_list):
    """Calculate the derivative of Rij vector w. r. t. atom m.
    
    Parameters
    ----------
    ijm_list: array [i*j, 3]
        List of indices of center atom i, neighbors atom j, and atom m.
    
    Returns
    -------
    list of float
        The derivative of the position vector R_{ij} with respect to atom
        index m in x, y, z directions.
    """

    dRij_dRm_vector = np.zeros(len(ijm_list))
    l1 = (ijm_list[:, 2] == ijm_list[:, 0])
    dRij_dRm_vector[l1] = -1
    l2 = (ijm_list[:, 2] == ijm_list[:, 1])
    dRij_dRm_vector[l2] = 1

    return dRij_dRm_vector


def dRijk_dRm(Rij, Rik, Rjk, ijk_list, m):
    """Calculate the derivative of R_{ab} norm with respect to atom m.
    
    Parameters
    ----------
    Rij: array [j*k, 3]
        The vector distances of atom i to neighbors js.
    Rik: array [j*k, 3]
        The vector distances of atom i to neighbors ks.
    Rjk: array [j*k, 3]
        The vector distances of atom j to atom k, where j != k.
    ijk_list: array
        The combinations of atom i, neighbor j, and neighbor k indices.
    m: array
        the list of atoms
        
    Returns
    -------
    dR{ab}_dRm: array [j*k, 3, m]
        The derivative of R_{ab} norm with respect to atom m.
    """
    dRij_dRm = np.zeros([len(Rij), 3, len(m)])
    dRik_dRm = np.zeros([len(Rij), 3, len(m)])
    dRjk_dRm = np.zeros([len(Rij), 3, len(m)])
    ij_list = ijk_list[:, [0, 1]]
    ik_list = ijk_list[:, [0, 2]]
    jk_list = ijk_list[:, [1, 2]]

    # for mm in m:
    for mm, _m in enumerate(m):
        mm_list = _m * np.ones([len(Rij), 1], dtype=int)
        dRij_dRm[:, :, mm] = dRij_dRm_norm(Rij, np.append(ij_list, mm_list,
                                                          axis=1))
        dRik_dRm[:, :, mm] = dRij_dRm_norm(Rik, np.append(ik_list, mm_list,
                                                          axis=1))
        dRjk_dRm[:, :, mm] = dRij_dRm_norm(Rjk, np.append(jk_list, mm_list,
                                                          axis=1))

    return (dRij_dRm, dRik_dRm, dRjk_dRm)


############################# Cutoff Functionals ##############################


def Cosine(Rij, Rc):
    # Rij is the norm 
    ids = (Rij > Rc)
    result = 0.5 * (np.cos(np.pi * Rij / Rc) + 1.)
    result[ids] = 0

    return result


def CosinePrime(Rij, Rc):
    # Rij is the norm
    ids = (Rij > Rc)
    result = -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)
    result[ids] = 0

    return result


########################### Auxiliary Functions ###############################


def to_array(input):
    if isinstance(input, (list, np.ndarray)):
        res = np.asarray(input)
    else:
        res = np.asarray([input])
    return res


def create_type_set(number_set, order=1):
    types = list(set(number_set))
    return np.array(list(combinations_with_replacement(types, order)))


def select_rows(data, row_pattern):
    if len(row_pattern) == 1:
        ids = (data == row_pattern)
    elif len(row_pattern) == 2:
        a, b = row_pattern
        ids = []
        for id, d in enumerate(data):
            if a == b:
                if d[0] == a and d[1] == a:
                    ids.append(id)
            else:
                if (d[0] == a and d[1] == b) or (d[0] == b and d[1] == a):
                    ids.append(id)
    return ids


################################### Test ######################################


if __name__ == '__main__':
    from ase.build import bulk

    np.set_printoptions(formatter={'float': '{: 0.4f}'.format})

    # Set up symmetry parameters
    Rc = 5.5
    symmetry = {'G2': {'eta': [0.036, 0.071, ], 'Rs': [0]},
                'G4': {'Rs': [0], 'lambda': [1], 'zeta': [1, ], 'eta': [0.036, 0.071]},
                'G5': {'Rs': [0], 'lambda': [1], 'zeta': [1, ], 'eta': [0.036, 0.071]},
                }

    for a in [5.0]:  # , 5.4, 5.8]:
        si = bulk('Si', 'diamond', a=a, cubic=True)
        cell = si.get_cell()
        # cell[0,1] += 0.2
        si.set_cell(cell)
        print(si.get_cell())

        bp = BehlerParrinello(symmetry, Rc=Rc, derivative=True, stress=True, atom_weighted=False)
        des = bp.calculate(si, system=[14])

        # print("G:", des['x'][0])
        # print("Sequence", des['seq'][0])
        # print("GPrime", des['dxdr'][0])
        # print("GPrime", des['dxdr'][0,:,2,:])
        # print("GPrime", des['dxdr'][0,:,4,:])
        # print(np.einsum('ijklm->klm', des['rdxdr']))

        bp = BehlerParrinello(symmetry, Rc=Rc, derivative=True, stress=True, atom_weighted=True)
        des = bp.calculate(si, system=[14])
        print("G:", des['x'][0])
        print("Sequence", des['seq'][0])
        print("GPrime", des['dxdr'].shape)
        # print("GPrime", des['dxdr'][0,:,0,:])
        # print("GPrime", des['dxdr'][0,:,2,:])
        # print("GPrime", des['dxdr'][0,:,4,:])



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn import init
from torch.nn.parameter import Parameter

try:

    from featurebox.layer.graph import _merge_cpp
    mod = _merge_cpp.mod
    me_idx_cpp = mod.merge_idx
    me_idx_mean_cpp = mod.merge_idx_mean
    me_idx_sum_cpp = mod.merge_idx_sum
    me_idx_max_cpp = mod.merge_idx_max
    me_idx_min_cpp = mod.merge_idx_min
except BaseException:
    mod = None

try:
    from torch_scatter import scatter_max
    from torch_scatter import scatter_add
    from torch_scatter import scatter_min
    from torch_scatter import scatter_mean

    scatter_mod = True
except ImportError:
    scatter_mod = None


class BaseLayer(nn.Module):
    """
    Base operation.
    """

    def __init__(self):
        super(BaseLayer, self).__init__()
        if mod is not None:
            self.merge_idx = self.merge_idx_cpp
        # elif scatter_mod is not None: # not very fast
        #     self.merge_idx = self.merge_idx_scatter
        else:
            self.merge_idx = self.merge_idx_py

    def merge_idx_methods(self, nbr_fea, node_atom_idx, methods=("mean", "max")):
        return torch.cat([self.merge_idx_py(nbr_fea, node_atom_idx, methodi) for methodi in methods], dim=-1)

    @staticmethod
    def merge_idx_py(nbr_fea, node_atom_idx, method="mean"):
        """
        Base realization.

        Parameters
        ----------
        nbr_fea:Tensor
        node_atom_idx:list
        method: callable,str

        Returns
        -------

        """
        # assert sum([len(idx_map) for idx_map in node_atom_idx]) == nbr_fea.data.shape[0]

        if method == "sum":
            temp = torch.cat([torch.sum(nbr_fea[i], dim=0, keepdim=True) for i in node_atom_idx])
        elif method == "max":
            temp = torch.cat([torch.max(nbr_fea[i], dim=0, keepdim=True)[1] for i in node_atom_idx])
        elif method == "min":
            temp = torch.cat([torch.min(nbr_fea[i], dim=0, keepdim=True)[1] for i in node_atom_idx])
        else:
            temp = torch.cat([torch.mean(nbr_fea[i], dim=0, keepdim=True) for i in node_atom_idx])

        return temp

    @staticmethod
    def merge_idx_cpp(nbr_fea, node_atom_idx, method="mean"):
        """Cpp realization,The first import is quite time consuming,
        but faster than *py realization in later call."""

        # assert sum([len(idx_map) for idx_map in node_atom_idx]) == nbr_fea.data.shape[0]
        return me_idx_cpp(nbr_fea, node_atom_idx, method)

    @staticmethod
    def merge_idx_cpp2(nbr_fea, node_atom_idx, method="mean"):
        """Cpp realization,the same with merge_idx_cpp,The first import is quite time consuming,
        but faster than *py realization in later call."""
        if method == "sum":
            temp = me_idx_sum_cpp(nbr_fea, node_atom_idx)
        elif method == "max":
            temp = me_idx_max_cpp(nbr_fea, node_atom_idx)
        elif method == "min":
            temp = me_idx_min_cpp(nbr_fea, node_atom_idx)
        else:
            temp = me_idx_mean_cpp(nbr_fea, node_atom_idx)

        return temp

    @staticmethod
    def merge_idx_scatter(nbr_fea, node_atom_idx, method="mean"):
        """torch-scatter realization,need torch-scatter model, it is useful in out-of-order index,
         but Not very fast in this part"""

        assert sum([len(idx_map) for idx_map in node_atom_idx]) == nbr_fea.data.shape[0]
        groups = torch.cat([torch.full((len(i),), n) for n, i in enumerate(node_atom_idx)])

        if method == "sum":
            temp = scatter_add(nbr_fea, groups, dim=0)
        elif method == "max":
            temp = scatter_max(nbr_fea, groups, dim=0)
        elif method == "min":
            temp = scatter_min(nbr_fea, groups, dim=0)
        else:
            temp = scatter_mean(nbr_fea, groups, dim=0)
        return temp

    @staticmethod
    def expand_idx(a, node_atom_idx):
        s2 = a.shape[1]
        lens = [len(i) for i in node_atom_idx]
        a = [ai.expand(i, s2) for i, ai in zip(lens, a)]
        return torch.cat(a, dim=0)


class MergeLayer(nn.Module):

    def forward(self, x_atom_fea, node_atom_idx):
        x_atom_fea_sum = self.pooling(x_atom_fea, node_atom_idx)
        return torch.mean(x_atom_fea_sum, dim=0, keepdim=True)

    def pooling(self, atom_fea, node_atom_idx):
        """
        Pooling the atom features to crystal features.\n
        N: Total number of atoms in the batch.\n
        N0: Total number of crystals in the batch.

        Parameters
        ----------
        atom_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom feature vectors of the batch
        node_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx
        """
        summed_fea = [torch.mean(atom_fea[idx_map], dim=0, keepdim=True)
                      for idx_map in node_atom_idx]
        return torch.cat(summed_fea, dim=0)


class UserLinear(nn.Module):
    __constants__ = ['max_in_features', 'out_features']

    max_in_features: int
    out_features: 1
    weight: Tensor

    def __init__(self,
                 max_in_features: int = 100, out_features: int = 20, bias: bool = True) -> None:
        super().__init__()
        self.max_in_features = max_in_features
        self.out_features = out_features
        self.weight = Parameter(torch.Tensor(out_features, max_in_features))
        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def forward(self, input: Tensor) -> Tensor:
        last_shape = input.shape[-1]
        assert last_shape < self.max_in_features, "The last axis size of input should be smaller tan max_in_features" \
                                                  "but they are {}>{}".format(last_shape, self.max_in_features)
        return F.linear(input, self.weight[..., :last_shape], self.bias[:last_shape])

    def extra_repr(self) -> str:
        return 'max_in_features={}, out_features={}, bias={}'.format(
            self.max_in_features, self.out_features, self.bias is not None
        )


class MergeLayerAgg(MergeLayer):
    def __init__(self, inner_number=20):
        super().__init__()

        self.tup = nn.ModuleList(
            (nn.Linear(1, inner_number),
             nn.Linear(2, inner_number),
             nn.Linear(3, inner_number),
             nn.Linear(4, inner_number),
             nn.Linear(5, inner_number),
             nn.Linear(6, inner_number),
             nn.Linear(7, inner_number),
             nn.Linear(8, inner_number),
             nn.Linear(9, inner_number),
             nn.Linear(10, inner_number),
             UserLinear(100, inner_number))
        )

    def forward(self, x_atom_fea, node_atom_idx):
        x_atom_fea_sum = self.pooling(x_atom_fea, node_atom_idx)
        x_atom_fea_sum2 = self.coef_pooling(x_atom_fea, node_atom_idx)
        x_atom_fea_sum = x_atom_fea_sum2 + x_atom_fea_sum
        return x_atom_fea_sum

    def coef_pooling(self, x_atom_fea, node_atom_idx):
        # x_atom_fea (N_atom,L_feature)
        # return (N_bond,L_feature)
        summed_fea = []
        for idx_map in node_atom_idx:
            if len(idx_map) - 1 <= 9:
                x = x_atom_fea[idx_map].T
                f = self.tup[len(idx_map) - 1]
                summed_feai = f(x).T

            else:
                x = x_atom_fea[idx_map]
                f = self.tup[-1]
                summed_feai = f(x)
            summed_feai = torch.sum(summed_feai, dim=0, keepdim=True)
            summed_fea.append(summed_feai)
        return torch.cat(summed_fea, dim=0)


if __name__ == "__main__":
    li = UserLinear(100)
    # index = [torch.arange(0,13),torch.arange(13,37)]
    a = torch.rand((37, 16)).T
    b = torch.rand((33, 16)).T
    a = a.to(torch.device("cuda:0"))
    li.to(torch.device("cuda:0"))
    resa = li(a)
    li.eval()
